# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines `CheckpointLayout`, a broader protocol used to save and load different checkpoint formats."""

from typing import Any, Awaitable, Protocol
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import types

Path = types.Path


class InvalidLayoutError(ValueError):
  """Raised when the checkpoint layout is invalid."""


class CheckpointLayout(Protocol):
  """CheckpointLayout.

  This class defines a protocol for different checkpoint formats. It is a helper
  component for `save_checkpointables` and
  `load_checkpointables`. It supports
  alternative checkpoint formats. It performs a few core functions:
    - Resolves handlers for saving and loading.
    - Saves and loads checkpointables to/from individual subdirectories by
    delegating to the resolved handlers.
  """

  @property
  def path(self) -> Path:
    """Returns the path of the checkpoint."""
    ...

  async def metadata(self) -> metadata_types.CheckpointMetadata[dict[str, Any]]:
    """Returns the metadata of the checkpoint.

    Returns:
      A dictionary of metadata. Dictionary keys represent the names of the
      checkpointables, while the values are the metadata objects themselves.
    """
    ...

  def validate(self) -> None:
    """Validates the path, determining if it conforms to this instance.

    Returns:
      None.

    Raises:
      InvalidLayoutError: If the path does not conform to this instance.
    """
    ...

  def validate_pytree(self, checkpointable_name: str | None) -> None:
    """Validates the path as a PyTree checkpoint.

    Args:
      checkpointable_name: The name of the checkpointable to load.

    Returns:
      None.

    Raises:
      InvalidLayoutError: If the path does not conform to this instance
        as a PyTree checkpoint.
    """
    ...

  async def load(
      self,
      abstract_checkpointables: dict[str, Any] | None = None,
  ) -> Awaitable[dict[str, Any]]:
    """Loads the checkpoint from the given directory.

    Args:
      abstract_checkpointables: A dictionary of abstract checkpointables.
        Dictionary keys represent the names of the checkpointables, while the
        values are the abstract checkpointable objects themselves.

    Returns:
      An awaitable dictionary of checkpointables. Dictionary keys represent the
      names of
      the checkpointables, while the values are the checkpointable objects
      themselves.
    """
    ...



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for validating checkpoint formats."""

from etils import epath
from orbax.checkpoint.experimental.v1._src.layout import checkpoint_layout
from orbax.checkpoint.experimental.v1._src.layout import orbax_layout
from orbax.checkpoint.experimental.v1._src.path import types as path_types

InvalidLayoutError = checkpoint_layout.InvalidLayoutError


def is_orbax_checkpoint(path: path_types.PathLike) -> bool:
  """Determines if the given path is an Orbax checkpoint.

  Args:
    path: The path to the checkpoint directory.

  Returns:
    True if the path is an Orbax checkpoint, False otherwise.
  """
  path = epath.Path(path)
  try:
    orbax_layout.OrbaxLayout(path).validate()
    return True
  except InvalidLayoutError:
    return False



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utilities for v0 checkpoint compatibility."""

from typing import Any

from orbax.checkpoint._src.checkpointers import async_checkpointer
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.handlers import handler_registration as legacy_handler_registration
from orbax.checkpoint._src.serialization import type_handlers
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import compatibility as handler_compatibility
from orbax.checkpoint.experimental.v1._src.handlers import composite_handler
from orbax.checkpoint.experimental.v1._src.loading import validation
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


PYTREE_CHECKPOINTABLE_KEY = format_utils.PYTREE_CHECKPOINTABLE_KEY
AbstractPyTree = tree_types.PyTreeOf[tree_types.AbstractLeafType]
CheckpointMetadata = metadata_types.CheckpointMetadata
PLACEHOLDER = type_handlers.PLACEHOLDER


def get_v0_checkpointer_and_args(
    path: path_types.Path,
    abstract_checkpointables: dict[str, Any] | None,
    *,
    context: context_lib.Context,
) -> tuple[
    async_checkpointer.AsyncCheckpointer,
    composite_checkpoint_handler.CompositeArgs,
]:
  """Construct V0 Checkpointer and Args for loading."""
  validation.validate_abstract_checkpointables(abstract_checkpointables)
  abstract_checkpointables = abstract_checkpointables or {}

  # pylint: disable=protected-access
  handlers = composite_handler.CompositeHandler(
      context.checkpointables_options.registry
  ).get_handlers_for_load(path, abstract_checkpointables)
  # pylint: enable=protected-access
  if not abstract_checkpointables:
    abstract_checkpointables = {
        name: None
        for name in handlers.keys()
        if name not in format_utils.RESERVED_CHECKPOINTABLE_KEYS
        and (path / name).exists()
    }

  compatibility_handlers = {
      name: handler_compatibility.get_compatibility_handler(handler)
      for name, handler in handlers.items()
  }
  legacy_handler_registry = (
      legacy_handler_registration.DefaultCheckpointHandlerRegistry()
  )
  for name, handler in compatibility_handlers.items():
    legacy_handler_registry.add(name, handler_compatibility.Args, handler)
  composite_options = composite_checkpoint_handler.CompositeOptions(
      async_options=context.async_options.v0(),
      file_options=context.file_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  ckptr = async_checkpointer.AsyncCheckpointer(
      composite_checkpoint_handler.CompositeCheckpointHandler(
          handler_registry=legacy_handler_registry,
          composite_options=composite_options,
      ),
      async_options=context.async_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      file_options=context.file_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  args = composite_checkpoint_handler.CompositeArgs(**{
      name: handler_compatibility.Args(checkpointable)
      for name, checkpointable in abstract_checkpointables.items()
  })
  return ckptr, args



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Internal utilities for saving whole and partial checkpoints."""

import time
from typing import Any, Awaitable
import uuid

from absl import logging
import jax
from orbax.checkpoint._src.futures import future
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.path import async_path
from orbax.checkpoint._src.path import atomicity
from orbax.checkpoint._src.path import atomicity_defaults
from orbax.checkpoint._src.path import atomicity_types
from orbax.checkpoint._src.path import utils as path_utils
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import composite_handler
from orbax.checkpoint.experimental.v1._src.handlers import types as handler_types
from orbax.checkpoint.experimental.v1._src.metadata import serialization as metadata_serialization
from orbax.checkpoint.experimental.v1._src.path import async_utils as path_async_utils
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.synchronization import thread_utils
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


InternalCheckpointMetadata = (
    step_metadata_serialization.InternalCheckpointMetadata
)


def get_temporary_path(
    path: path_types.Path, *, context: context_lib.Context
) -> atomicity_types.TemporaryPath:
  """Gets a TemporaryPath for the given path."""
  temporary_path_class = (
      context.file_options.temporary_path_class
      or atomicity_defaults.get_default_temporary_path_class(path)
  )
  tmpdir = temporary_path_class.from_final(
      path,
      # Ensure metadata store is NOT passed, to prevent separate metadata
      # writing.
      checkpoint_metadata_store=None,
      file_options=context.file_options.v0(),
  )
  return tmpdir


async def remove_existing_path(
    path: path_types.Path,
    *,
    context: context_lib.Context,
):
  if multihost.is_primary_host(context.multiprocessing_options.primary_host):
    logging.info(
        '[process=%s] Specified `overwrite`: removing existing path.',
        multihost.process_index(),
    )
    await async_path.rmtree(path)
  await multihost.sync_global_processes(
      multihost.unique_barrier_key(
          'save_checkpointables_async:rmtree',
          prefix=context.multiprocessing_options.barrier_sync_key_prefix,
      ),
      processes=context.multiprocessing_options.active_processes,
  )


def add_internal_checkpointables(
    checkpointables: dict[str, Any],
    *,
    context: context_lib.Context,
    metrics: tree_types.JsonType | None = None,
) -> dict[str, Any]:
  """Adds descriptor to checkpointables if enabled."""
  # Global registration ties metrics key to JsonHandler.
  if metrics:
    checkpointables[format_utils.METRICS_CHECKPOINTABLE_KEY] = metrics
  return checkpointables


def record_save_start(path: path_types.Path, *, async_origin: bool):
  """Records the start of a save operation."""
  logging.info(
      '[process=%s] Started %s checkpoint to %s.',
      multihost.process_index(),
      'async saving' if async_origin else 'saving',
      path,
  )
  if async_origin:
    event_name = '/jax/orbax/write/async/start'
  else:
    event_name = '/jax/orbax/write/start'
  jax.monitoring.record_event(event_name)
  jax.monitoring.record_event(
      '/jax/orbax/write/storage_type',
      storage_type=path_utils.get_storage_type(path),
  )


def record_save_completion(
    path: path_types.Path,
    *,
    total_duration_secs: float,
    async_origin: bool,
):
  """Records the completion of a save operation."""
  logging.info(
      'Finished asynchronous save (blocking + background) in %.2f seconds'
      ' to %s',
      total_duration_secs,
      path,
  )
  # TODO(cpgaffney): No event is currently being recorded for synchronous saves.
  # Consider collecting this information
  if async_origin:
    jax.monitoring.record_event_duration_secs(
        '/jax/checkpoint/write/async/total_duration_secs',
        total_duration_secs,
    )


class SaveResponse(async_types.AsyncResponse[None]):
  """An `AsyncResponse` representing the result of `save_pytree_async`."""

  def __init__(
      self,
      operation_id: str,
      tmp_path: atomicity_types.TemporaryPath,
      handler_typestrs: dict[str, str],
      background_awaitable: Awaitable[None],
      *,
      start_time: float,
      custom_metadata: tree_types.JsonType | None,
      context: context_lib.Context,
      async_origin: bool,
  ):
    self._operation_id = operation_id
    self._tmp_path = tmp_path
    self._handler_typestrs = handler_typestrs
    self._background_awaitable = background_awaitable
    self._start_time = start_time
    self._custom_metadata = custom_metadata
    self._context = context
    self._async_origin = async_origin
    self._thread_runner = thread_utils.BackgroundThreadRunner[None](
        self._finalize_save()
    )

  async def _finalize_save(self):
    logging.info(
        '[process=%s] Finalizing checkpoint on %s',
        multihost.process_index(),
        self._tmp_path.get(),
    )
    await self._background_awaitable
    logging.vlog(
        1,
        '[process=%s] Finished waiting for background save operations.',
        multihost.process_index(),
    )

    if multihost.is_primary_host(
        self._context.multiprocessing_options.primary_host
    ):
      logging.vlog(
          1,
          '[process=%s] Writing checkpoint metadata.',
          multihost.process_index(),
      )
      internal_metadata = InternalCheckpointMetadata.create(
          handler_typestrs=self._handler_typestrs,
          init_timestamp_nsecs=int(self._start_time * 1e9),
          commit_timestamp_nsecs=time.time_ns(),
          custom_metadata=self._custom_metadata,
      )
      await metadata_serialization.write(
          metadata_serialization.checkpoint_metadata_file_path(
              self._tmp_path.get()
          ),
          internal_metadata.serialize(),
      )
      logging.vlog(
          1,
          '[process=%s] Finished writing checkpoint metadata.',
          multihost.process_index(),
      )
      await atomicity.on_commit_callback(
          self._tmp_path,
          checkpoint_start_time=self._start_time,
      )

    # Clean up all awaitable signals for the current operation id as they are
    # no longer needed.
    if self._context.async_options.create_directories_asynchronously:
      future.AwaitableSignalsContract.remove_all_awaitable_signals(
          self._operation_id
      )

    await multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'save_checkpointables_async:finalize',
            prefix=self._context.multiprocessing_options.barrier_sync_key_prefix,
        ),
        processes=self._context.multiprocessing_options.active_processes,
    )
    total_duration_secs = time.time() - self._start_time
    record_save_completion(
        self._tmp_path.get_final(),
        total_duration_secs=total_duration_secs,
        async_origin=self._async_origin,
    )

  def result(self, timeout: float | None = None) -> None:
    return self._thread_runner.result(timeout=timeout)


async def maybe_overwrite_existing(
    path: path_types.Path,
    *,
    overwrite: bool,
    context: context_lib.Context,
) -> None:
  """Checks if the path exists and overwrites it if necessary."""
  if await async_path.exists(path):
    if overwrite:
      await remove_existing_path(path, context=context)
    else:
      raise ValueError(f'Destination {path} already exists.')


async def run_blocking_save(
    tmp_path: atomicity_types.TemporaryPath,
    checkpointables: dict[str, Any],
    *,
    overwrite: bool,
    context: context_lib.Context,
) -> Awaitable[None]:
  """Runs the synchronous portion of the save operation.

  This includes directory creation and calling the handler's save method.

  Args:
    tmp_path: The temporary path to save the checkpointables to.
    checkpointables: A mapping from checkpointable name to checkpointable.
    overwrite: Whether to overwrite an existing checkpoint in `tmp_path`.
    context: The context to use for the save operation.

  Returns:
    An awaitable that will be completed when the synchronous portion of the
    save operation is complete.
  """
  await context_lib.synchronize_next_operation_id()

  await maybe_overwrite_existing(
      tmp_path.get_final(), overwrite=overwrite, context=context
  )

  handler = composite_handler.CompositeHandler(
      context.checkpointables_options.registry
  )

  # Directory creation is handled here.
  tmp_path_awaiting_creation = path_async_utils.start_async_mkdir(
      tmp_path, checkpointables.keys()
  )
  if not context.async_options.create_directories_asynchronously:
    await tmp_path_awaiting_creation.await_creation()

  # Delegate to the handler to get the background awaitable.
  background_awaitable = await handler.save(
      tmp_path_awaiting_creation, checkpointables
  )
  return background_awaitable


def create_save_response(
    background_awaitable: Awaitable[None],
    checkpointables: dict[str, Any],
    tmp_path: atomicity_types.TemporaryPath,
    start_time: float,
    *,
    context: context_lib.Context,
    custom_metadata: tree_types.JsonType | None,
    async_origin: bool,
) -> async_types.AsyncResponse[None]:
  """Creates and returns the final AsyncResponse for a save operation."""
  blocking_duration_secs = time.time() - start_time
  jax.monitoring.record_event_duration_secs(
      '/jax/checkpoint/write/async/blocking_duration_secs',
      blocking_duration_secs,
  )
  logging.info(
      'Finished blocking save in %.2f seconds. Continuing to write to %s.',
      blocking_duration_secs,
      tmp_path.get_final(),
  )

  handler = composite_handler.CompositeHandler(
      context.checkpointables_options.registry
  )
  handler_typestrs = {
      name: handler_types.typestr(type(handler))
      for name, handler in handler.get_handlers_for_save(
          checkpointables
      ).items()
  }

  return SaveResponse(
      context.operation_id(),
      tmp_path,
      handler_typestrs,
      background_awaitable,
      start_time=start_time,
      custom_metadata=custom_metadata,
      context=context,
      async_origin=async_origin,
  )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Public symbols for checkpoint_managers module."""

# pylint: disable=g-importing-member, g-bad-import-order, unused-import, g-multiple-import

from orbax.checkpoint._src.checkpoint_managers import save_decision_policy
from orbax.checkpoint._src.checkpoint_managers.save_decision_policy import (
    DecisionContext,
    SaveDecisionPolicy,
    FixedIntervalPolicy,
    InitialSavePolicy,
    SpecificStepsPolicy,
    ContinuousCheckpointingPolicy,
    AnySavePolicy,
)

from orbax.checkpoint._src.checkpoint_managers import preservation_policy
from orbax.checkpoint._src.checkpoint_managers.preservation_policy import (
    PreservationContext,
    PreservationPolicy,
    LatestN,
    EveryNSeconds,
    EveryNSteps,
    CustomSteps,
    AnyPreservationPolicy,
    BestN,
)

from orbax.checkpoint._src.checkpoint_managers.policy_checkpoint_info import PolicyCheckpointInfo
from orbax.checkpoint.checkpoint_manager import CheckpointManagerOptions
from orbax.checkpoint.checkpoint_manager import CheckpointManager

from orbax.checkpoint.abstract_checkpoint_manager import AbstractCheckpointManager



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Orbax step storage entities."""

import abc
from collections.abc import Iterable
from concurrent import futures
import dataclasses
import datetime
import functools
import os
import re
import time
from typing import Callable, Generic, Iterator, List, Optional, Protocol, Sequence, Set, TypeVar
from absl import logging
from etils import epath
import numpy as np
from orbax.checkpoint._src.metadata import checkpoint
from orbax.checkpoint._src.metadata import step_metadata_serialization
from orbax.checkpoint._src.multihost import multihost
from orbax.checkpoint._src.path import gcs_utils


_GCS_PATH_PREFIX = ('gs://',)
_COMMIT_SUCCESS_FILE = 'commit_success.txt'
TMP_DIR_SUFFIX = '.orbax-checkpoint-tmp'
# prefix_1000.orbax-checkpoint-tmp-1010101
# OR
# 1000.orbax-checkpoint-tmp-1010101
TMP_DIR_STEP_PATTERN = r'.*?_*?(\d+)\.orbax-checkpoint-tmp'

MetadataT = TypeVar('MetadataT', bound='Metadata')


@dataclasses.dataclass(frozen=True)
class Metadata:
  """Metadata of a step.

  Attributes:
    step: step number of the checkpoint.
    path: path to the checkpoint.
  """

  step: int
  path: epath.Path

  @functools.cached_property
  def _checkpoint_metadata(self) -> Optional[checkpoint.StepMetadata]:
    """Returns checkpoint metadata of this step."""
    metadata_dict = checkpoint.metadata_store(enable_write=False).read(
        file_path=checkpoint.step_metadata_file_path(self.path)
    )
    if metadata_dict is None:
      return None
    return step_metadata_serialization.deserialize(metadata_dict)

  @property
  def init_timestamp_nsecs(self) -> Optional[int]:
    """Returns init timestamp of uncommitted checkpoint of this step.

    It is specified as nano seconds since epoch.
    """
    metadata = self._checkpoint_metadata
    if metadata is None:
      return None
    return metadata.init_timestamp_nsecs

  @property
  def commit_timestamp_nsecs(self) -> Optional[int]:
    """Returns commit timestamp of the checkpoint of this step.

    It is specified as nano seconds since epoch.
    """
    metadata = self._checkpoint_metadata
    if metadata is None:
      return None
    return metadata.commit_timestamp_nsecs

  @property
  def commit_timestamp(self) -> datetime.datetime:
    """Returns commit timestamp of the checkpoint of this step.

    It is specified as datetime in UTC timezone.
    """
    commit_timestamp_nsecs = self.commit_timestamp_nsecs
    if commit_timestamp_nsecs is not None:
      timestamp_sec = commit_timestamp_nsecs / 1e9
    else:
      timestamp_sec = self.path.stat().mtime
    return datetime.datetime.fromtimestamp(
        timestamp_sec, tz=datetime.timezone.utc
    )


class NameFormat(Protocol, Generic[MetadataT]):
  """Protocol responsible for naming and querying steps."""

  def build_name(self, step: int) -> str:
    """Returns `step` name.

    *Implementation hint:* Implement it to build a name for the given step using
    the class's custom formatting attributes. Since it is mainly meant for
    building names to save checkpoints, it can raise error if this NameFormat is
    just meant for finding already existing step paths.

    Args:
      step: Step number.
    """
    ...

  @abc.abstractmethod
  def find_all(self, base_path: epath.PathLike) -> Iterator[MetadataT]:
    """Returns metadata of all steps.

    NOTE: Ignores uncommitted checkpoints.

    *Implementation hint:* Implement it to find all step folders under
    `base_path` performing IO operations if needed. Use
    `build_step_metadatas(...)` helper function to build all the `MetadataT`
    using the found step paths.

    Args:
      base_path: *root* Path under which Step folders are placed.
    """
    ...

  @abc.abstractmethod
  def find_step(self, base_path: epath.PathLike, step: int) -> MetadataT:
    """Returns the metadata for `step` or raises ValueError.

    NOTE: Ignores uncommitted checkpoints.

    *Implementation hint:* Implement it to find the step folder under
    `base_path` performing IO operations if needed.

    Args:
      base_path: *root* Path under which Step folders are placed.
      step: Step number.

    Raises:
      ValueError if no committed paths for the requested step is found.
    """
    ...


def build_step_path(
    base_path: epath.PathLike, name_format: NameFormat[Metadata], step: int
) -> epath.Path:
  """Returns `step` path under `base_path` for step `name_format`."""
  return epath.Path(base_path) / name_format.build_name(step)


def build_step_metadatas(
    step_paths: Iterable[epath.Path],
    build_metadata: Callable[[epath.Path], Optional[MetadataT]],
) -> Iterator[MetadataT]:
  """Yields filtered metadata mapped with `step_paths`.

  Args:
    step_paths: Iterator of step paths.
    build_metadata: Callable to match and build step metadata from `step_paths`
      elements. If a `step_paths` element doesn't match then it returns None.

  Yields:
    Step metadata.
  """
  with futures.ThreadPoolExecutor() as executor:
    metadata_futures = [
        executor.submit(build_metadata, step_path) for step_path in step_paths
    ]
    for future in futures.as_completed(metadata_futures):
      metadata = future.result()
      if metadata is not None:
        yield metadata


def step_prefix_with_underscore(step_prefix: Optional[str]) -> str:
  """Returns `step_prefix` appended with `underscore` or <empty> if None."""
  return '' if step_prefix is None else f'{step_prefix}_'


def latest_step_metadata(
    root_path: epath.PathLike, name_format: NameFormat[MetadataT]
) -> Optional[MetadataT]:
  """Returns step.MetadataT of the latest step in `root_path`."""
  return max(
      sorted(
          name_format.find_all(root_path),
          key=lambda metadata: metadata.path.name,
          reverse=True,
      ),
      default=None,
      key=lambda metadata: metadata.step,
  )


def step_metadata_of_checkpoint_path(
    checkpoint_path: epath.PathLike, name_format: NameFormat[MetadataT]
) -> MetadataT:
  """Returns step.MetadataT of given `checkpoint_path`."""
  checkpoint_path = epath.Path(checkpoint_path)
  all_step_metadata = list(name_format.find_all(checkpoint_path.parent))
  for step_metadata in all_step_metadata:
    if step_metadata.path.name == checkpoint_path.name:
      return step_metadata
  raise ValueError(
      'Failed to resolve step metadata of checkpoint path with'
      f' NameFormat={name_format}, checkpoint path={checkpoint_path}, path'
      f' name({checkpoint_path.name}) did not match with available step names:'
      f' {[step_metadata.path.name for step_metadata in all_step_metadata]}.'
      ' Please check if the given path is really a checkpoint path.'
  )


# TODO(b/337858698): Works with CompositeNameFormat.write_name_format only. Also
# support read_name_formats.
def find_step_path(
    base_path: epath.PathLike,
    name_format: NameFormat[Metadata],
    *,
    step: int,
    include_uncommitted: bool = False,
) -> epath.Path:
  """Returns `step` path under `base_path` for step `name_format`.

  NOTE: Experimental function, subject to change.

  Args:
    base_path: directory path containing step subdirs.
    name_format: NameFormat of the target `step`.
    step: target step number.
    include_uncommitted: if True then uncommitted steps are considered in search
      too, otherwise only committed steps are looked up.

  Raises:
    ValueError if the target step path does not exist.
  """
  base_path = epath.Path(base_path)
  if not include_uncommitted:
    return name_format.find_step(base_path, step).path

  # First try finding uncommitted step.
  if is_gcs_path(base_path):
    uncommitted_step_path = build_step_path(base_path, name_format, step)
  else:
    step_name = name_format.build_name(step)
    uncommitted_step_path = None
    for uncommitted_name in tmp_checkpoints(base_path):
      if uncommitted_name.startswith(f'{step_name}{TMP_DIR_SUFFIX}'):
        uncommitted_step_path = base_path / uncommitted_name
        break
  if uncommitted_step_path and uncommitted_step_path.exists():
    return uncommitted_step_path
  # Uncommitted step not found, return committed one or raise error.
  return name_format.find_step(base_path, step).path


def maybe_find_step_metadata(
    base_path: epath.PathLike,
    name_format: NameFormat[Metadata],
    *,
    step: int,
) -> Metadata | None:
  """Returns `Metadata` for `step` with `name_format` or None."""
  try:
    return name_format.find_step(base_path, step)
  except ValueError:
    return None


@dataclasses.dataclass(frozen=True)
class _StandardNameFormat(NameFormat[Metadata]):
  """NameFormat for 'standard' steps for common Orbax use cases.

  NOTE: Ignores uncommitted checkpoints.

  Naming examples:
   * step_prefix=None    step_format_fixed_length=None  ->  23
   * step_prefix=None    step_format_fixed_length=4     ->  0023
   * step_prefix=step    step_format_fixed_length=None  ->  step_23
   * step_prefix=step    step_format_fixed_length=4     ->  step_0023

  Attributes:
    step_prefix: Optional fixed string prefixed to step. Note an *underscore* is
      appended before applying it.
    step_format_fixed_length: Optional length of the zero padded step. e.g. 6
      for 000123.
    single_host_load_and_broadcast: If True, the jax process=0 will list all
      steps and broadcast them to all other processes. NOTE: Ignored if jax
      backend is not multi controller.
    enable_hns: Enables HNS-specific path logic.
  """

  step_prefix: Optional[str] = None
  step_format_fixed_length: Optional[int] = None
  single_host_load_and_broadcast: bool = False
  enable_hns: bool = False

  def __str__(self):
    return f'StandardNameFormat("{self.build_name(1234)}")'

  def build_name(self, step: int) -> str:
    """Returns `(prefix_)?(zero padding)?step` name."""
    if self.step_format_fixed_length is not None:
      step_str = f'{step:0{self.step_format_fixed_length}d}'
    else:
      step_str = f'{step}'

    # [prefix]step
    return f'{step_prefix_with_underscore(self.step_prefix)}{step_str}'

  def _build_metadata(
      self, step_path: epath.Path, step: Optional[int] = None
  ) -> Optional[Metadata]:
    """Returns metadata for given `step_path` if it is valid or None."""
    if not _is_checkpoint_finalized_ignore_error(step_path):
      return None

    if step is not None:
      # step already known, just check exists.
      if step_path.exists():
        return Metadata(step=step, path=step_path)

    # Regex: [prefix]*(step)
    if self.step_format_fixed_length and self.step_format_fixed_length > 0:
      zero_present = rf'0\d{{{self.step_format_fixed_length-1}}}'
      zero_not_present = rf'[1-9]\d{{{self.step_format_fixed_length-1}}}\d*'
      zero_padded_step_group = rf'({zero_present}|{zero_not_present})'
    else:
      zero_padded_step_group = r'(0|[1-9]\d*)'
    name_regex = f'^{step_prefix_with_underscore(self.step_prefix)}{zero_padded_step_group}$'

    match = re.search(name_regex, step_path.name)
    if match is None:
      return None
    (step_,) = match.groups()
    step_ = int(step_)

    return Metadata(step=step_, path=step_path)

  def _glob_step_paths(self, base_path: epath.PathLike) -> list[epath.Path]:
    """Returns step paths under `base_path`."""
    base_path = epath.Path(base_path)
    # <step_prefix>_?<0 padding>?*
    if self.enable_hns and gcs_utils.is_hierarchical_namespace_enabled(
        base_path
    ):
      logging.vlog(
          1,
          'HNS enabled. Using GCS API to list step paths at %s',
          base_path.as_posix(),
      )
      bucket_name, path_prefix = gcs_utils.parse_gcs_path(base_path)
      bucket = gcs_utils.get_bucket(bucket_name)
      result = bucket.list_blobs(
          prefix=path_prefix,
          delimiter='/',
          include_folders_as_prefixes=True,
      )
      # Iterate over pages to force a fetch from the server, after which
      # `result.prefixes` will be populated.
      for _ in result.pages:
        pass
      return [
          epath.Path(f'gs://{bucket_name}/{folder}')
          for folder in result.prefixes
          if folder.startswith(os.path.join(path_prefix, self.step_prefix))
      ]
    else:
      return list(
          epath.Path(base_path).glob(
              f'{step_prefix_with_underscore(self.step_prefix)}*'
          )
      )

  def _find_all_with_single_host_load_and_broadcast(
      self, base_path: epath.PathLike
  ) -> Iterator[Metadata]:
    """Returns metadata by validating and broadcasting from process_index=0.

    Even though step_paths are gathered on all hosts, the validation and
    extraction of step numbers of complete checkpoints is done only on the
    process_index=0 (this involves multiple file I/O operations and arguably
    improves performance by reducing QPS to underlying storages like GCS). The
    step numbers are then broadcast to all hosts. In the end, all hosts build
    the metadata from step numbers and return.

    Args:
      base_path: Root path under which step folders are placed.
    """
    process_index = multihost.process_index()
    time_start = time.time()
    # <step_prefix>_?<0 padding>?*
    step_paths = self._glob_step_paths(base_path)
    if not step_paths:
      logging.info(
          '[process=%s][single_host_load_and_broadcast] No steps found,'
          ' root_dir=%s',
          process_index,
          base_path,
      )
      return iter([])
    time_glob_list = time.time()
    base_path = epath.Path(base_path)
    max_steps = len(step_paths)
    padded_step_list = np.array([-1] * max_steps)
    # Read the step list only on the primary host, and then broadcast the list.
    # This minimizes queries to underlying storages like GCS when number of
    # steps is large.
    is_primary_host = process_index == 0
    if is_primary_host:
      steps = []
      with futures.ThreadPoolExecutor() as executor:
        metadata_futures = [
            executor.submit(self._build_metadata, step_path)  # File IO
            for step_path in step_paths
        ]
        for future in futures.as_completed(metadata_futures):
          metadata = future.result()
          if metadata is not None:
            steps.append(metadata.step)
      steps.sort()
      steps = np.array(steps)
      assert len(steps) <= max_steps
      padded_step_list[0 : len(steps)] = steps
    padded_step_list = multihost.broadcast_one_to_all(
        padded_step_list, is_source=is_primary_host
    )
    paths_to_step_dict: dict[epath.Path, int] = {
        base_path / self.build_name(step): step
        for step in padded_step_list
        if step >= 0
    }
    metadatas = build_step_metadatas(
        paths_to_step_dict.keys(),
        lambda step_path: Metadata(
            step=paths_to_step_dict[step_path], path=step_path
        ),
    )
    time_build_metadata = time.time()
    logging.info(
        '[process=%s][single_host_load_and_broadcast] Found #steps=%s,'
        ' root_dir=%s, time_taken_to_glob=%ss,'
        ' time_taken_to_build_metadata=%ss, total_time=%ss',
        process_index,
        len(paths_to_step_dict),
        base_path,
        time_glob_list - time_start,
        time_build_metadata - time_glob_list,
        time_build_metadata - time_start,
    )
    return metadatas

  def find_all(self, base_path: epath.PathLike) -> Iterator[Metadata]:
    """Returns metadata of all steps matching with name_format attributes."""
    if multihost.process_count() > 1 and self.single_host_load_and_broadcast:
      return self._find_all_with_single_host_load_and_broadcast(base_path)

    # <step_prefix>_?<0 padding>?*
    step_paths = self._glob_step_paths(base_path)
    return build_step_metadatas(step_paths, self._build_metadata)

  def find_step(self, base_path: epath.PathLike, step: int) -> Metadata:
    """Returns the metadata for `step` or raises ValueError."""
    step_path = build_step_path(base_path, self, step)
    metadata = self._build_metadata(step_path, step=step)
    if metadata is not None:
      return metadata

    # Raise detailed error message.
    raise ValueError(
        f'No step path found with name={self.build_name(step)},'
        f' NameFormat={self} for step={step} under {base_path}.'
    )


def standard_name_format(
    *,
    step_prefix: Optional[str] = None,
    step_format_fixed_length: Optional[int] = None,
    single_host_load_and_broadcast: bool = False,
    enable_hns: bool = False,
) -> NameFormat[Metadata]:
  """Returns NameFormat for 'standard' steps for common Orbax use cases.

  NOTE: Ignores uncommitted checkpoints.

  Naming examples:
   * step_prefix=None    step_format_fixed_length=None  ->  23
   * step_prefix=None    step_format_fixed_length=4     ->  0023
   * step_prefix=step    step_format_fixed_length=None  ->  step_23
   * step_prefix=step    step_format_fixed_length=4     ->  step_0023

  Args:
    step_prefix: Optional fixed string prefixed to step. Note an *underscore* is
      appended before applying it.
    step_format_fixed_length: Optional length of the zero padded step. e.g. 6
      for 000123.
    single_host_load_and_broadcast: If True, the jax process=0 will list all
      steps and broadcast them to all other processes. NOTE: Ignored if jax
      backend is not multi controller.
    enable_hns: Enables HNS-specific path logic.
  """
  return _StandardNameFormat(
      step_prefix=step_prefix,
      step_format_fixed_length=step_format_fixed_length,
      single_host_load_and_broadcast=single_host_load_and_broadcast,
      enable_hns=enable_hns,
  )


@dataclasses.dataclass(frozen=True)
class _CompositeNameFormat(NameFormat[Metadata]):
  """Supports reading multiple step namings, but just one format to write.

  Attributes:
    write_name_format: NameFormat used to build step names meant for writing
      checkpoints. Must be present in `read_name_formats` at a preferred
      priority position.
    read_name_formats: Sequence (ordered) of NameFormats used to find steps for
      reading checkpoints. It acts like an *or*, where the first one to match is
      returned.
  """

  write_name_format: NameFormat[Metadata]
  read_name_formats: Sequence[NameFormat[Metadata]]

  def __post_init__(self):
    if self.write_name_format not in self.read_name_formats:
      read_name_formats = ','.join(str(f) for f in self.read_name_formats)
      raise ValueError(
          f'write_name_format: {self.write_name_format} must be present in'
          f' read_name_formats: [{read_name_formats}].'
      )

  def __str__(self):
    read_name_formats = ','.join(str(f) for f in self.read_name_formats)
    return f'Composite([{read_name_formats}])'

  def build_name(self, step: int) -> str:
    """Returns `step` name using `write_name_format`."""
    return self.write_name_format.build_name(step)

  def find_all(self, base_path: epath.PathLike) -> Iterator[Metadata]:
    """Returns metadata of all steps."""
    found_paths = set()
    for read_name_format in self.read_name_formats:
      for step_metadata in read_name_format.find_all(base_path):
        if step_metadata.path not in found_paths:
          found_paths.add(step_metadata.path)
          yield step_metadata

  def find_step(self, base_path: epath.PathLike, step: int) -> Metadata:
    """Returns the metadata for `step` or raises ValueError."""
    errors = []  # Used to raise the final collated error if needed.
    for read_name_format in self.read_name_formats:
      try:
        return read_name_format.find_step(base_path, step)
      except Exception as e:  # pylint: disable=broad-exception-caught
        logging.info(
            'Failed to find step=%s with NameFormat=%s under %s. Error: %s',
            step,
            read_name_format,
            base_path,
            e,
        )
        errors.append(e)

    # Raise the concatenated errors.
    messages = [f'{e}' for e in errors]
    raise ValueError('\n'.join(messages))


def composite_name_format(
    write_name_format: NameFormat[Metadata],
    read_name_formats: Sequence[NameFormat[Metadata]],
) -> NameFormat[Metadata]:
  """Returns *composite* NameFormat supporting multiple read/single write formats.

  Args:
    write_name_format: NameFormat used to build step names meant for writing
      checkpoints. Must be present in `read_name_formats` at a preferred
      priority position.
    read_name_formats: Sequence (ordered) of NameFormats used to find steps for
      reading checkpoints. Please note that to resolve conflicts (and avoid
      raising errors) in case of multiple NameFormats matching a given step, the
      sequence should be provided in highest to lowest priority order:
      NameFormat appearing earlier in the sequence is preferred.
  """
  return _CompositeNameFormat(write_name_format, read_name_formats)


# TODO(b/337137764) Can't move it to path/utils due to cyclic dependency.
# Explore other options.
def is_gcs_path(path: epath.Path) -> bool:
  return path.as_posix().startswith(_GCS_PATH_PREFIX)


# TODO(b/337137764) Can't move it to path/utils due to cyclic dependency.
# Explore other options.
def get_save_directory(
    step: int,
    directory: epath.PathLike,
    name: Optional[str] = None,
    step_prefix: Optional[str] = None,
    override_directory: Optional[epath.PathLike] = None,
    step_format_fixed_length: Optional[int] = None,
    step_name_format: Optional[NameFormat[Metadata]] = None,
) -> epath.Path:
  """Returns the standardized path to a save directory for a single item.

  Args:
    step: Step number.
    directory: Top level checkpoint directory.
    name: Item name ('params', 'state', 'dataset', etc.).
    step_prefix: Prefix applied to `step` (e.g. 'checkpoint').
    override_directory: If provided, step, directory, and step_prefix are
      ignored.
    step_format_fixed_length: Uses a fixed number of digits with leading zeros
      to represent the step number. If None, there are no leading zeros.
    step_name_format: NameFormat used to define step name for step and under
      given root directory. If provided, `step_prefix` and
      `step_format_fixed_length` are ignored.

  Returns:
    A directory.
  """
  if directory is None:
    raise ValueError('Directory cannot be None.')
  directory = epath.Path(directory)
  if override_directory is not None:
    result = epath.Path(override_directory)
  else:
    step_name_format = step_name_format or standard_name_format(
        step_prefix=step_prefix,
        step_format_fixed_length=step_format_fixed_length,
    )
    result = build_step_path(directory, step_name_format, step)
  if name is not None:
    result /= name
  return result


def is_tmp_checkpoint(path: epath.PathLike) -> bool:
  """Determines whether a directory is a tmp checkpoint path."""
  path = epath.Path(path)
  if os.path.islink(path):
    logging.warning(
        'Path %s is a symbolic link, not treating it as a tmp checkpoint.', path
    )
    return False
  if not path.exists():
    raise ValueError(f'Path {path} does not exist.')
  if not path.is_dir():
    return False
  if is_gcs_path(path) and not (path / _COMMIT_SUCCESS_FILE).exists():
    return True
  if TMP_DIR_SUFFIX in path.name:
    return True
  return False


def is_checkpoint_finalized(path: epath.PathLike) -> bool:
  """Determines if the given path represents a finalized checkpoint.

  Path takes the form::

    path/to/my/dir/<name>.orbax-checkpoint-tmp-<timestamp>/  # not finalized
    path/to/my/dir/<name>/  # finalized

  Alternatively::

    gs://path/to/my/dir/<name>/  # finalized
      commit_success.txt
      ...
    gs://<path/to/my/dir/<name>/  # not finalized
      ...

  Args:
    path: Directory.

  Returns:
    True if the checkpoint is finalized.

  Raises:
    ValueError if the provided path is not a directory. Valid checkpoint paths
    must be a directory.
  """
  path = epath.Path(path)

  # Keep the following line as the first condition.
  # is_gcs_path is inmemory and fast to bypass when dealing with non-gcs fs
  if is_gcs_path(path) and not (path / _COMMIT_SUCCESS_FILE).exists():
    logging.warning(
        'This GCS path %s does not contain the %s file used to indicate a'
        ' successfully written GCS checkpoint. If the checkpoint was'
        ' originally saved with GCS, the checkpoint was not successfully'
        ' written. If the checkpoint was saved differently and copied, you'
        ' need to add %s to the checkpoint directory.',
        path,
        _COMMIT_SUCCESS_FILE,
        _COMMIT_SUCCESS_FILE,
    )

    return False
  if not path.exists():
    raise ValueError(f'Path {path} does not exist.')
  if not path.is_dir():
    raise ValueError(f'Path {path} is not a directory. Not a valid checkpoint')
  if TMP_DIR_SUFFIX in path.name:
    return False
  return True


def _is_checkpoint_finalized_ignore_error(path: epath.PathLike) -> bool:
  try:
    return is_checkpoint_finalized(path)
  except ValueError:
    return False


def tmp_checkpoints(checkpoint_dir: epath.PathLike) -> List[str]:
  """Returns a list of tmp checkpoint dir names in `checkpoint_dir`."""
  checkpoint_dir = epath.Path(checkpoint_dir)
  if not checkpoint_dir.exists():
    return []
  return [s.name for s in checkpoint_dir.iterdir() if is_tmp_checkpoint(s)]


def cleanup_tmp_directories(
    directory: epath.PathLike,
    primary_host: Optional[int] = 0,
    active_processes: Optional[Set[int]] = None,
    barrier_sync_key_prefix: Optional[str] = None,
):
  """Cleanup steps in `directory` with tmp files, as these are not finalized."""
  directory = epath.Path(directory)
  if multihost.is_primary_host(primary_host):
    logging.info('Cleaning up existing temporary directories at %s.', directory)
    tmp_files = tmp_checkpoints(directory)
    for tmp_file in tmp_files:
      (directory / tmp_file).rmtree()
  multihost.sync_global_processes(
      multihost.unique_barrier_key(
          'cleanup_tmp_dirs',
          prefix=barrier_sync_key_prefix,
      ),
      timeout=multihost.coordination_timeout(),
      processes=active_processes,
  )


def _is_legacy_step_checkpoint(path: epath.Path) -> bool:
  """Determines if the path resembles an Orbax step directory.

  Note that this is not foolproof, and users should not add extra files to the
  checkpoint directory beyond what is done by CheckpointManager.

  Args:
    path: path to check.

  Returns:
    bool indicating whether the path resembles an Orbax step directory.
  """
  name = path.name
  # Path must be a directory and either a digit, or end in '_' + digit.
  return path.is_dir() and (name.isdigit() or name.split('_')[-1].isdigit())


def _is_legacy_finalized_step_checkpoint(path: epath.Path) -> bool:
  return _is_legacy_step_checkpoint(
      path
  ) and _is_checkpoint_finalized_ignore_error(path)


def step_from_checkpoint_name(name: str) -> int:
  """Returns the step from a checkpoint name. Also works for tmp checkpoints."""
  if name.isdigit():
    return int(name)
  elif name.split('_')[-1].isdigit():
    split = name.split('_')
    if len(split) == 2 and split[0]:
      return int(split[-1])
  elif tmp_match := re.match(TMP_DIR_STEP_PATTERN, name):
    return int(tmp_match.group(1))
  raise ValueError(f'Unrecognized name format: {name}.')


def checkpoint_steps_paths(
    checkpoint_dir: epath.PathLike,
) -> List[epath.Path]:
  """Returns a list of finalized checkpoint paths in the directory."""
  checkpoint_dir = epath.Path(checkpoint_dir)
  if not checkpoint_dir.exists():
    return []

  with futures.ThreadPoolExecutor() as executor:
    fs = {
        step_dir: executor.submit(
            _is_legacy_finalized_step_checkpoint, step_dir
        )
        for step_dir in checkpoint_dir.iterdir()
    }
    return [step_dir for step_dir, future in fs.items() if future.result()]


def checkpoint_steps(
    checkpoint_dir: epath.PathLike, single_host_load_and_broadcast: bool = False
) -> List[int]:
  """Returns a list of finalized checkpoint steps in the directory."""

  def _checkpoint_steps(path: epath.Path) -> List[int]:
    return [
        step_from_checkpoint_name(s.name) for s in checkpoint_steps_paths(path)
    ]

  checkpoint_dir = epath.Path(checkpoint_dir)
  if single_host_load_and_broadcast:
    max_steps = len(list(checkpoint_dir.iterdir()))
    # Read the step list only from host 0, and then broadcast the list.
    # This minimizes queries on non-leader processes.
    padded_step_list = np.array([-1] * max_steps)
    if multihost.process_index() == 0:
      steps = np.array(_checkpoint_steps(checkpoint_dir))
      assert len(steps) <= max_steps
      padded_step_list[0 : len(steps)] = steps
    padded_step_list = multihost.broadcast_one_to_all(padded_step_list)
    return [step for step in padded_step_list if step >= 0]
  return _checkpoint_steps(checkpoint_dir)


def any_checkpoint_step(checkpoint_dir: epath.PathLike) -> Optional[int]:
  """Returns any finalized checkpoint step in the directory or None.

  This avoids iterating over the entire directory.

  Args:
    checkpoint_dir: Checkpoint directory.

  Returns:
    Any finalized checkpoint step in the directory or None.
  """
  checkpoint_dir = epath.Path(checkpoint_dir)
  for s in checkpoint_dir.iterdir():
    if _is_legacy_finalized_step_checkpoint(s):
      return step_from_checkpoint_name(s.name)
  return None



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Checkpoint deleter."""

import datetime
import os
import pathlib
import queue
import threading
import time
from typing import Optional, Protocol, Sequence
from absl import logging
from etils import epath
import jax
from orbax.checkpoint import utils
from orbax.checkpoint._src.path import gcs_utils
from orbax.checkpoint._src.path import step as step_lib
PurePosixPath = pathlib.PurePosixPath

_THREADED_DELETE_DURATION = (
    '/jax/orbax/checkpoint_manager/threaded_checkpoint_deleter/duration'
)

_STANDARD_DELETE_DURATION = (
    '/jax/orbax/checkpoint_manager/standard_checkpoint_deleter/duration'
)


class CheckpointDeleter(Protocol):
  """A protocol defined a CheckpointDeleter."""

  def delete(self, step: int) -> None:
    """Delete a step."""
    ...

  def delete_steps(self, steps: Sequence[int]) -> None:
    """Delete steps."""
    ...

  def close(self) -> None:
    """Perform any cleanup before closing this deleter."""
    ...


class StandardCheckpointDeleter:
  """A StandardCheckpointDeleter."""

  def __init__(
      self,
      directory: epath.Path,
      *,
      name_format: step_lib.NameFormat[step_lib.Metadata],
      primary_host: Optional[int] = 0,
      todelete_subdir: Optional[str] = None,
      todelete_full_path: Optional[str] = None,
      enable_hns: bool = False,
      duration_metric: Optional[str] = _STANDARD_DELETE_DURATION,
  ):
    """StandardCheckpointDeleter constructor.

    Args:
      directory: refer to CheckpointManager.directory
      name_format: refer to CheckpointManager._name_format
      primary_host: refer to CheckpointManager.primary_host
      todelete_subdir: refer to CheckpointManagerOptions.todelete_subdir
      todelete_full_path: refer to CheckpointManagerOptions.todelete_full_path
      enable_hns: refer to CheckpointManagerOptions.enable_hns
      duration_metric: the name of the total delete duration metric
    """
    self._primary_host = primary_host
    self._directory = directory
    self._todelete_subdir = todelete_subdir
    self._todelete_full_path = todelete_full_path
    self._name_format = name_format
    self._enable_hns = enable_hns
    self._duration_metric = duration_metric

  def _rm_empty_folders(self, path: epath.Path) -> None:
    """For a hierarchical namespace bucket, delete empty folders recursively."""
    # pylint: disable=g-import-not-at-top
    from google.cloud import storage_control_v2  # pytype: disable=import-error

    bucket, prefix = gcs_utils.parse_gcs_path(path)

    client = storage_control_v2.StorageControlClient()
    project_path = client.common_project_path('_')
    bucket_path = f'{project_path}/buckets/{bucket}'
    folders = set(
        # Format: "projects/{project}/buckets/{bucket}/folders/{folder}"
        folder.name
        for folder in client.list_folders(
            request=storage_control_v2.ListFoldersRequest(
                parent=bucket_path, prefix=prefix.strip('/') + '/'
            )
        )
    )

    while folders:
      parents = set(os.path.dirname(x.rstrip('/')) + '/' for x in folders)
      leaves = folders - parents
      requests = [
          storage_control_v2.DeleteFolderRequest(name=f) for f in leaves
      ]
      for req in requests:
        client.delete_folder(request=req)
      folders = folders - leaves
      logging.vlog(
          1,
          'Deleted %s folders, %s remaining. [%s][%s]',
          len(leaves),
          len(folders),
          bucket,
          prefix,
      )

  def _rmtree(self, path: epath.Path):
    """Recursively deletes a path.

    For a hierarchical namespace bucket, `path.rmtree()` only removes objects,
    leaving all the empty parent folders intact. Here we manually delete the
    empty folders recursively.

    Args:
      path: the path to delete.
    """
    # Step 1: Delete all files within the tree.
    path.rmtree()

    # Step 2: For HNS, clean up the remaining empty directory structure.
    if self._enable_hns:
      if gcs_utils.is_hierarchical_namespace_enabled(path):
        self._rm_empty_folders(path)

  def delete(self, step: int) -> None:
    """Deletes step dir or renames it if options are set.

    See `CheckpointManagerOptions.todelete_subdir` for details.

    Args:
      step: checkpointing step number.
    """
    start = time.time()
    try:
      if not utils.is_primary_host(self._primary_host):
        logging.info(
            'Not primary host(%s), skipping deletion of step %d.',
            self._primary_host,
            step,
        )
        return

      try:
        delete_target = step_lib.find_step_path(
            self._directory,
            self._name_format,
            step=step,
            include_uncommitted=True,
        )
      except ValueError as e:
        logging.warning(
            'Unable to find the step %d for deletion or renaming, err=%s',
            step,
            e,
        )
        return

      # Attempt to rename using GCS HNS API if configured.
      if self._todelete_full_path is not None:
        if step_lib.is_gcs_path(self._directory):
          self._rename_gcs_step_with_hns(step, delete_target)
          return
        else:
          raise NotImplementedError()

      # Attempt to rename to local subdirectory using `todelete_subdir`
      # if configured.
      if self._todelete_subdir is not None and not step_lib.is_gcs_path(
          self._directory
      ):
        self._rename_step_to_subdir(step, delete_target)
        return

      # The final case: fall back to permanent deletion.
      self._delete_step_permanently(step, delete_target)

    finally:
      jax.monitoring.record_event_duration_secs(
          self._duration_metric,
          time.time() - start,
      )

  def _rename_gcs_step_with_hns(
      self, step: int, delete_target: epath.Path
  ):
    """Renames a GCS directory using the Storage Control API.

    Args:
      step: The checkpoint step number.
      delete_target: The path to the directory to be renamed.

    Raises:
      ValueError: If the GCS bucket is not HNS-enabled, as this is a
        hard requirement for this operation.
    """
    logging.info(
        'Condition: GCS path with `todelete_full_path` set. Checking for HNS.'
    )
    bucket_name, _ = gcs_utils.parse_gcs_path(self._directory)
    if not gcs_utils.is_hierarchical_namespace_enabled(self._directory):
      raise ValueError(
          f'Bucket "{bucket_name}" does not have Hierarchical Namespace'
          ' enabled, which is required when _todelete_full_path is set.'
      )

    logging.info('HNS bucket detected. Attempting to rename step %d.', step)
    # pylint: disable=g-import-not-at-top
    from google.api_core import exceptions as google_exceptions  # pytype: disable=import-error
    try:
      from google.cloud import storage_control_v2  # pytype: disable=import-error
      import google.auth  # pytype: disable=import-error

      # Use default credentials, but without a quota project to avoid
      # quota issues with this API.
      credentials, _ = google.auth.default()
      creds_without_quota_project = credentials.with_quota_project(None)
      client = storage_control_v2.StorageControlClient(
          credentials=creds_without_quota_project
      )
      # Destination parent is the absolute path to the bucket.
      destination_parent_dir_str = (
          f'gs://{bucket_name}/{self._todelete_full_path}'
      )
      destination_parent_path = PurePosixPath(destination_parent_dir_str)
      logging.info(
          'Ensuring destination parent folder exists via HNS API: %s',
          destination_parent_dir_str,
      )
      try:
        parent_folder_id = str(
            destination_parent_path.relative_to(f'gs://{bucket_name}')
        )
        bucket_resource_name = f'projects/_/buckets/{bucket_name}'
        client.create_folder(
            request=storage_control_v2.CreateFolderRequest(
                parent=bucket_resource_name,
                folder_id=parent_folder_id,
                recursive=True,
            )
        )
        logging.info('HNS parent folder creation request sent.')
      except google_exceptions.AlreadyExists:
        logging.info('HNS parent folder already exists, proceeding.')

      now = datetime.datetime.now()
      timestamp_str = now.strftime('%Y%m%d-%H%M%S-%f')
      new_name_with_timestamp = f'{delete_target.name}-{timestamp_str}'
      dest_path = destination_parent_path / new_name_with_timestamp
      source_folder_id = str(delete_target.relative_to(f'gs://{bucket_name}'))
      destination_folder_id = str(dest_path.relative_to(f'gs://{bucket_name}'))
      source_resource_name = (
          f'projects/_/buckets/{bucket_name}/folders/{source_folder_id}'
      )
      logging.info('Rename API call: Source: %s', source_resource_name)
      logging.info('Rename API call: Destination ID: %s', destination_folder_id)
      request = storage_control_v2.RenameFolderRequest(
          name=source_resource_name,
          destination_folder_id=destination_folder_id,
      )
      op = client.rename_folder(request=request)
      op.result()
      logging.info('Successfully renamed step %d to %s', step, dest_path)
    except google_exceptions.GoogleAPIError as e:
      logging.error('HNS rename failed for step %d. Error: %s', step, e)

  def _rename_step_to_subdir(self, step: int, delete_target: epath.Path):
    """Renames a step directory to its corresponding todelete_subdir."""
    rename_dir = self._directory / self._todelete_subdir
    rename_dir.mkdir(parents=True, exist_ok=True)
    dst = step_lib.build_step_path(rename_dir, self._name_format, step)
    delete_target.replace(dst)
    logging.info('Renamed step %d to %s', step, dst)

  def _delete_step_permanently(self, step: int, delete_target: epath.Path):
    """Permanently deletes a step directory."""
    self._rmtree(delete_target)
    logging.info('Deleted step %d.', step)

  def delete_steps(self, steps: Sequence[int]) -> None:
    for step in steps:
      self.delete(step)

  def close(self) -> None:
    pass


class ThreadedCheckpointDeleter:
  """A threaded CheckpointDeleter."""

  def __init__(
      self,
      directory: epath.Path,
      *,
      name_format: step_lib.NameFormat[step_lib.Metadata],
      primary_host: Optional[int] = 0,
      todelete_subdir: Optional[str] = None,
      todelete_full_path: Optional[str] = None,
      enable_hns: bool = False,
  ):
    """ThreadedCheckpointDeleter deletes checkpoints in a background thread."""
    self._standard_deleter = StandardCheckpointDeleter(
        primary_host=primary_host,
        directory=directory,
        todelete_subdir=todelete_subdir,
        todelete_full_path=todelete_full_path,
        name_format=name_format,
        enable_hns=enable_hns,
        duration_metric=_THREADED_DELETE_DURATION,
    )
    self._delete_queue = queue.Queue()
    # Turn on daemon=True so the thread won't block the main thread and die
    # when the program exits.
    self._delete_thread = threading.Thread(
        target=self._delete_thread_run, name='DeleterThread', daemon=True
    )
    self._delete_thread.start()

    jax.monitoring.record_event(
        '/jax/orbax/checkpoint_manager/threaded_checkpoint_deleter/init'
    )

  def _delete_thread_run(self) -> None:
    logging.info('Delete thread has started.')
    while True:
      step = self._delete_queue.get(block=True)
      if step < 0:
        break
      self._standard_deleter.delete(step)
    logging.info('Delete thread exited.')

  def delete(self, step: int) -> None:
    self._delete_queue.put(step)

  def delete_steps(self, steps: Sequence[int]) -> None:
    for step in steps:
      self.delete(step)

  def close(self) -> None:
    # make sure all steps are deleted before exit.
    if self._delete_thread and self._delete_thread.is_alive():
      self._delete_queue.put(-1)
      self._delete_thread.join()

    self._standard_deleter.close()


def create_checkpoint_deleter(
    directory: epath.Path,
    *,
    name_format: step_lib.NameFormat[step_lib.Metadata],
    primary_host: Optional[int] = 0,
    todelete_subdir: Optional[str] = None,
    todelete_full_path: Optional[str] = None,
    enable_hns: bool = False,
    enable_background_delete: bool = False,
) -> CheckpointDeleter:
  """Creates a CheckpointDeleter."""

  if enable_background_delete:
    return ThreadedCheckpointDeleter(
        directory,
        name_format=name_format,
        primary_host=primary_host,
        todelete_subdir=todelete_subdir,
        todelete_full_path=todelete_full_path,
        enable_hns=enable_hns,
    )
  else:
    return StandardCheckpointDeleter(
        directory,
        name_format=name_format,
        primary_host=primary_host,
        todelete_subdir=todelete_subdir,
        todelete_full_path=todelete_full_path,
        enable_hns=enable_hns,
    )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for constructing maximal shardings."""

from __future__ import annotations
import math
from typing import Any
from absl import logging
import jax
import numpy as np

PyTree = Any


def _partition_axis_name(offset: int) -> str:
  return str(chr(ord('a') + offset))


def _construct_maximal_sharding(
    sds: jax.ShapeDtypeStruct,
) -> jax.sharding.Sharding:
  """Constructs a sharding that partitions the array as much as possible."""
  shape = sds.shape
  if not shape:
    return jax.sharding.NamedSharding(
        mesh=jax.sharding.Mesh(jax.devices(), ('a',)),
        spec=jax.sharding.PartitionSpec(),
    )
  if np.max(shape) < jax.device_count():
    # Array is small - no sharding needed.
    return jax.sharding.NamedSharding(
        mesh=jax.sharding.Mesh(jax.devices(), ('a',)),
        spec=jax.sharding.PartitionSpec(),
    )

  available_device_dim = jax.device_count()
  partition_axes = [None] * len(shape)
  mesh_shape = []
  mesh_axes = []

  current_parition_axis = 0
  # Max to min.
  for i in np.argsort(shape)[::-1]:
    assert available_device_dim > 0
    if available_device_dim == 1:
      break
    if shape[i] < available_device_dim:
      continue
    gcd = math.gcd(shape[i], available_device_dim)
    if gcd == 1:
      continue
    available_device_dim //= gcd
    mesh_shape.append(gcd)

    current_parition_axis_name = _partition_axis_name(current_parition_axis)
    partition_axes[i] = current_parition_axis_name
    mesh_axes.append(current_parition_axis_name)
    current_parition_axis += 1

  # Still have some partition dimension left over.
  if available_device_dim > 1:
    mesh_shape.append(available_device_dim)
    mesh_axes.append(_partition_axis_name(current_parition_axis))

  logging.info(
      'Constructed sharding for array with shape: %s, mesh_shape: %s,'
      ' mesh_axes: %s, partition_axes: %s',
      shape,
      mesh_shape,
      mesh_axes,
      partition_axes,
  )

  assert len(mesh_shape) == len(mesh_axes)
  assert len(partition_axes) == len(shape)
  mesh = jax.sharding.Mesh(
      np.asarray(jax.devices()).reshape(mesh_shape),
      tuple(mesh_axes),
  )
  pspec = jax.sharding.PartitionSpec(*partition_axes)
  return jax.sharding.NamedSharding(mesh=mesh, spec=pspec)


def construct_maximal_shardings(abstract_state: PyTree) -> PyTree:
  """Construct a sharding that partitions each array as much as possible.

  This method is subject to change and should not be considered stable.

  Args:
    abstract_state: PyTree of jax.ShapeDtypeStruct.

  Returns:
    PyTree of jax.sharding.Sharding.
  """
  shardings = jax.tree.map(_construct_maximal_sharding, abstract_state)

  total_size = 0

  def _calculate_sharding_hbm_consumption(
      sds: jax.ShapeDtypeStruct, sharding: jax.sharding.Sharding
  ):
    nonlocal total_size
    shard_shape = sharding.shard_shape(sds.shape)
    total_size += np.prod(shard_shape) * sds.dtype.itemsize

  jax.tree.map(_calculate_sharding_hbm_consumption, abstract_state, shardings)
  logging.info('Expected per-device HBM consumption: %s', total_size)
  return shardings



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Synchronization utilities for futures."""

import enum
import itertools
import threading
import time
from typing import Callable, Generic, TypeVar
from absl import logging
from orbax.checkpoint import options as options_lib
from orbax.checkpoint._src.multihost import multihost

_T = TypeVar("_T")


class HandlerAwaitableSignal(enum.Enum):
  """Defines signals that may be awaited by a `CheckpointHandler`.

  Signals may be passed from `CheckpointManager` or `Checkpointer` layers to
  `CheckpointHandler or below.`

  Attributes:
    AWAITABLE_SIGNALS_CONTRACT: Contract that contains a list of signals that
      may be sent and can be awaited by the handlers.
    STEP_DIRECTORY_CREATION: When recieved, indicates that the step directory
      has been created. The handler should not attempt to write files before the
      directory is created.
    ITEM_DIRECTORY_CREATION: When recieved, indicates that the item directory
      has been created. The handler should not attempt to write files before the
      directory is created.
  """

  AWAITABLE_SIGNALS_CONTRACT = "awaitable_signals_contract"
  STEP_DIRECTORY_CREATION = "step_directory_creation"
  ITEM_DIRECTORY_CREATION = "item_directory_creation"


class OperationIdGenerator:
  """A unique operation id generator."""

  _operation_id_counter = itertools.count()
  _operation_id = next(_operation_id_counter)

  @classmethod
  def next_operation_id(cls) -> int:
    """Increments the operation id counter and returns the new value."""
    cls._operation_id = next(cls._operation_id_counter)
    return cls._operation_id

  @classmethod
  def get_current_operation_id(cls) -> str:
    """Returns the current operation id."""
    return str(cls._operation_id)


class MultihostSynchronizedValue(Generic[_T]):
  """A thread-safe value that is synchronized across all processes."""

  def __init__(
      self,
      value: _T,
      multiprocessing_options: options_lib.MultiprocessingOptions,
      async_options: options_lib.AsyncOptions,
  ):
    """Initializes the thread-safe value."""
    self._multiprocessing_options = multiprocessing_options
    self._async_options = async_options
    self._lock = threading.RLock()
    with self._lock:
      self._value = value

  def _create_thread_safe_barrier_sync_fn(self) -> Callable[[str], None]:
    """Returns a barrier sync function to be called from threads.

    The function accepts a key, but the timeout is already set up using
    `AsyncOptions.timeout_secs` attribute.

    The Jax based barrier sync util, `sync_global_devices`, should not be called
    concurrently. Otherwise, it may cause a deadlock.

    In general, any Jax function with `collectives` should not be called
    concurrently to avoid deadlocks.
    """
    async_options = self._async_options or options_lib.AsyncOptions()
    timeout_secs = async_options.timeout_secs
    barrier_sync_fn = (
        async_options.barrier_sync_fn
        or multihost.get_barrier_sync_fn(
            processes=self._multiprocessing_options.active_processes
        )
    )
    return lambda key: barrier_sync_fn(key=key, timeout_ms=timeout_secs * 1000)

  def get(self) -> _T:
    """Returns the value."""
    with self._lock:
      return self._value

  def set(self, value: _T) -> None:
    """Sets the value across all processes."""
    start_time = time.time()
    thread_safe_barrier_sync_fn = self._create_thread_safe_barrier_sync_fn()
    thread_safe_barrier_sync_fn(
        multihost.unique_barrier_key(
            "ThreadSaveMultiHostValueHolder:set_value_start",
            prefix=self._multiprocessing_options.barrier_sync_key_prefix,
        ),
    )
    with self._lock:
      self._value = value
      thread_safe_barrier_sync_fn(
          multihost.unique_barrier_key(
              "ThreadSaveMultiHostValueHolder:set_value_end",
              prefix=self._multiprocessing_options.barrier_sync_key_prefix,
          ),
      )
    logging.vlog(
        1,
        "[process=%s]MultihostSynchronizedValue set took %s seconds",
        multihost.process_index(),
        time.time() - start_time,
    )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for path constructs."""

import asyncio
from collections.abc import AsyncIterable
import os
import time
from typing import Iterable, Optional

from absl import logging
from etils import epath



def get_storage_type(path: epath.Path | str) -> str:
    return 'other'


class Timer(object):
  """A simple timer to measure the time it takes to run a function."""

  def __init__(self):
    self._start = time.time()

  def get_duration(self):
    return time.time() - self._start




def recursively_copy_files(
    src: epath.PathLike,
    dst: epath.PathLike,
    *,
    skip_paths: Iterable[str] | None = None,
) -> None:
  """Recursively copies files from src to dst.

  Args:
    src: The source directory to copy from.
    dst: The destination directory to copy to.
    skip_paths: An optional iterable of relative paths to skip.
  """
  src_path = epath.Path(src)
  dst_path = epath.Path(dst)
  skip_paths_set = set(skip_paths) if skip_paths is not None else set()

  for root, dirs, files in os.walk(src_path):
    relative_path = str(root)[len(str(src_path)) :].lstrip(os.sep)
    if relative_path in skip_paths_set:
      continue
    # Prune dirs that are in skip_paths to prevent traversal.
    dirs[:] = [
        d for d in dirs if os.path.join(relative_path, d) not in skip_paths_set
    ]

    dst_root = dst_path / relative_path
    dst_root.mkdir(parents=True, exist_ok=True)

    for file in files:
      relative_file_path = os.path.join(relative_path, file)
      if relative_file_path in skip_paths_set:
        continue

      src_file = epath.Path(root) / file
      dst_file = epath.Path(dst_root) / file
      src_file.copy(dst_file)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Async wrappers for path operations.

Any function conforming to::

  epath.Path.foo(*args, **kwargs)

Should be implemented as::

  async def foo(path: epath.Path, *args, **kwargs):
    return await asyncio.to_thread(path.foo, *args, **kwargs)
"""

import asyncio
from typing import Any, Iterator

from etils import epath



async def mkdir(
    path: epath.Path,
    parents: bool = False,
    exist_ok: bool = False,
    mode: int | None = None,
):
  """Creates a directory asynchronously."""

  def _mkdir_sync(**thread_kwargs):
    """Synchronously creates a directory."""
    path.mkdir(parents=parents, exist_ok=exist_ok, mode=mode)

  thread_kwargs = {}
  return await asyncio.to_thread(_mkdir_sync, **thread_kwargs)


async def write_bytes(path: epath.Path, data: Any) -> int:
  return await asyncio.to_thread(path.write_bytes, data)


async def read_bytes(path: epath.Path) -> bytes:
  return await asyncio.to_thread(path.read_bytes)


async def write_text(path: epath.Path, text: str) -> int:
  return await asyncio.to_thread(path.write_text, text)


async def read_text(path: epath.Path) -> str:
  return await asyncio.to_thread(path.read_text)


async def exists(path: epath.Path):
  return await asyncio.to_thread(path.exists)


async def async_stat(path: epath.Path):
  return await asyncio.to_thread(path.stat)


async def async_rmtree(path: epath.Path):
  return await asyncio.to_thread(path.rmtree)


async def rmtree(path: epath.Path):
  return await asyncio.to_thread(path.rmtree)


async def touch(path: epath.Path, *, exist_ok: bool = False):
  return await asyncio.to_thread(path.touch, exist_ok=exist_ok)


async def rename(src: epath.Path, dst: epath.Path):
  return await asyncio.to_thread(src.rename, dst)


async def glob(path: epath.Path, pattern: str) -> Iterator[epath.Path]:
  return await asyncio.to_thread(path.glob, pattern)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Orbax utils related to multihost_utils functionality."""

import threading
import time
from typing import List, Optional, Protocol, Set
from absl import flags
from absl import logging
import jax
from jax.experimental import multihost_utils
import numpy as np

# Default timeout in seconds.
_DEFAULT_BARRIER_TIMEOUT = 300

# Used in unit tests with multiple parallel test cases.
_TEST_CASE_INDEX = None

# Map from distributed process index to device ids.
_DISTRIBUTED_TO_DEVICE_IDS: List[List[int]] = None

# Map from runtime process index to distributed process index.
_RUNTIME_TO_DISTRIBUTED_ID: List[int] = None

EXPERIMENTAL_ORBAX_USE_DISTRIBUTED_PROCESS_ID = flags.DEFINE_bool(
    'experimental_orbax_use_distributed_process_id',
    False,
    'If True, uses jax._src.distributed.global_state.process_id instead of'
    ' jax.process_index().',
)


def coordination_timeout() -> int:
  """Returns the timeout for coordination operations."""
  return _DEFAULT_BARRIER_TIMEOUT




def is_runtime_to_distributed_ids_initialized() -> bool:
  return _RUNTIME_TO_DISTRIBUTED_ID is not None


def is_distributed_to_device_ids_initialized() -> bool:
  return _DISTRIBUTED_TO_DEVICE_IDS is not None


def distributed_to_device_ids() -> List[List[int]]:
  if not is_distributed_to_device_ids_initialized():
    raise ValueError('Please call initialize_distributed_to_device_ids().')
  return _DISTRIBUTED_TO_DEVICE_IDS


def initialize_distributed_to_device_ids():
  """Initializes the device index mapping."""
  # Already initialized.
  if is_distributed_to_device_ids_initialized():
    return
  global _DISTRIBUTED_TO_DEVICE_IDS

  client = get_jax_distributed_client()
  own_distributed_id = jax._src.distributed.global_state.process_id  # pylint: disable=protected-access
  own_device_ids = [d.id for d in jax.local_devices()]
  dir_key = 'orbax/distributed_to_device_ids/'
  key = dir_key + str(own_distributed_id)
  client.key_value_set(key, str(own_device_ids))
  client.wait_at_barrier(
      # 5 minutes.
      'orbax_global_discovery_device_ids',
      timeout_in_ms=5 * 60 * 1000,
  )
  ids = client.key_value_dir_get(dir_key)
  results = [None for _ in range(jax.process_count())]
  for key, device_ids in ids:
    distributed_id = int(key.split('/')[-1])
    # Remove the list brackets.
    device_ids = device_ids[1 : len(device_ids) - 1]
    results[distributed_id] = [
        int(device_id) for device_id in device_ids.split(', ')
    ]
  assert None not in results
  _DISTRIBUTED_TO_DEVICE_IDS = results
  logging.vlog(
      1,
      '[process=%s][thread=%s] distributed_to_device_ids: %s',
      own_distributed_id,
      threading.current_thread().name,
      _DISTRIBUTED_TO_DEVICE_IDS,
  )


# TODO(cpgaffney): Delete this method.
def initialize_runtime_to_distributed_ids():
  """Initializes the process index mapping.

  Note that this function is for experimental purposes only. It is intended to
  be called exactly once per processes.

  TODO(b/325293150): Remove once the bug is resolved.
  """
  global _RUNTIME_TO_DISTRIBUTED_ID
  client = get_jax_distributed_client()

  # Index is distributed id.
  # Value is runtime id.
  _RUNTIME_TO_DISTRIBUTED_ID = [0 for _ in range(jax.process_count())]
  own_runtime_id = jax.process_index()
  own_distributed_id = jax._src.distributed.global_state.process_id  # pylint: disable=protected-access
  dir_key = 'jax/process_id/'
  key = dir_key + str(own_runtime_id)
  client.key_value_set(key, str(own_distributed_id))
  client.wait_at_barrier('orbax_global_discovery', timeout_in_ms=5 * 60 * 1000)
  ids = client.key_value_dir_get(dir_key)
  for key, distributed_id in ids:
    runtime_id = int(key.split('/')[-1])
    _RUNTIME_TO_DISTRIBUTED_ID[runtime_id] = int(distributed_id)
  logging.vlog(
      1,
      '[process=%s][thread=%s] runtime_to_distributed_id: %s',
      process_index(),
      threading.current_thread().name,
      _RUNTIME_TO_DISTRIBUTED_ID,
  )


def runtime_to_distributed_ids() -> List[int]:
  if not is_runtime_to_distributed_ids_initialized():
    raise ValueError('Please call initialize_runtime_to_distributed_ids().')
  return _RUNTIME_TO_DISTRIBUTED_ID


def runtime_to_distributed_process_id(pid: int) -> int:
  """Converts a distributed process index to a runtime process index."""
  if not is_runtime_to_distributed_ids_initialized():
    raise ValueError('Please call initialize_runtime_to_distributed_ids().')
  return _RUNTIME_TO_DISTRIBUTED_ID[pid]


def broadcast_one_to_all(in_tree, is_source: Optional[bool] = None):
  """Broadcast data from a source host to all other hosts."""
  if is_source is None:
    is_source = process_index() == 0
  return multihost_utils.broadcast_one_to_all(in_tree, is_source=is_source)


def should_skip_process_sync(processes: Optional[Set[int]] = None) -> bool:
  if processes and len(processes) == 1 and process_index() in processes:
    return True
  if jax.process_count() == 1:
    return True
  return False


def is_jax_distributed_client_initialized() -> bool:
  """Returns True if the JAX distributed client is initialized."""
  return jax._src.distributed.global_state.client is not None  # pylint: disable=protected-access


def get_jax_distributed_client():
  client = jax._src.distributed.global_state.client  # pylint: disable=protected-access
  if client is None:
    raise ValueError(
        'Distributed system is not available; please initialize it via '
        '`jax.distributed.initialize()` at the start of your program.'
    )
  return client


class BarrierSyncFn(Protocol):
  """Protocol for a barrier synchronization callable."""

  def __call__(self, *, key: str, timeout_ms: int) -> None:
    """Blocks on a barrier identified by key with the given timeout."""
    ...


def get_barrier_sync_fn(
    *,
    processes: Optional[Set[int]] = None,
) -> BarrierSyncFn:
  """Provides a barrier synchronization function for JAX processes.

  Barriers with different sync keys are safe to use from independent background
  threads.

  Args:
    processes: If None, expects to wait across all processes and devices.
      Otherwise, creates a barrier only across devices associated with the given
      processes.

  Returns:
    A no-op function if there is a single JAX process, or
    A barrier synchronization callable which accepts two arguments: "key": [str]
    unique barrier id; "timeout_ms": [int] timeout to use for waiting on the
    barrier.
    Should be called from all JAX processes with the same sync key and will
    block until either 1) all processes have reached the barrier or
    2) the timeout is exceeded.
  """
  if jax.process_count() == 1:
    return lambda **kwargs: None

  client = get_jax_distributed_client()
  barrier_processes = processes or set(range(jax.process_count()))
  if process_index() not in barrier_processes:
    raise ValueError(
        'Attempted to create a barrier across a subset of processes, but the'
        f' current process: {process_index()} was not present in the provided'
        f' list of processes: {barrier_processes}.'
    )

  # Use distributed ids.
  if processes is None:
    barrier_processes = None
  else:
    # Don't map ids anymore if we are using distributed ids.
    barrier_processes = list(barrier_processes)

  def _fn(*, key: str, timeout_ms: int) -> None:
    key = _unique_barrier_key(key)
    logging.vlog(
        1,
        '[process=%s][thread=%s] Waiting at barrier: %s',
        process_index(),
        threading.current_thread().name,
        key,
    )
    if processes is None:
      client.wait_at_barrier(key, timeout_ms)
    else:
      logging.vlog(
          1,
          '[process=%s][thread=%s] Barrier processes: %s',
          process_index(),
          threading.current_thread().name,
          barrier_processes,
      )
      client.wait_at_barrier(key, timeout_ms, process_ids=barrier_processes)
    logging.vlog(
        1,
        '[process=%s][thread=%s] Done waiting at barrier: %s',
        process_index(),
        threading.current_thread().name,
        key,
    )

  return _fn


def _unique_barrier_key(key: str) -> str:
  """Function that can be overridden for testing purposes."""
  return key


def unique_barrier_key(
    key: str,
    *,
    prefix: Optional[str] = None,
    suffix: Optional[str] = None,
) -> str:
  """Constructs a key given an optional prefix and suffix."""
  if prefix is not None:
    key = f'{prefix}_{key}'
  if suffix is not None:
    key = f'{key}.{suffix}'
  return key


def sync_global_processes(
    name: str,
    *,
    timeout: Optional[int] = None,
    processes: Optional[Set[int]] = None,
    barrier_sync_fn: Optional[BarrierSyncFn] = None,
    record_event_name: str = '/jax/checkpoint/sync_global_devices_duration_sec',
):
  """Barrier to sync concurrent processes.

  NOTE: The barrier name must be unique, i.e. no process should wait on the
  same barrier name multiple times.

  Args:
    name: barrier name. Must be unique.
    timeout: timeout in seconds.
    processes: If None, expects to wait across all processes and devices.
      Otherwise, creates a barrier only across devices associated with the given
      processes.
    barrier_sync_fn: Used as the implementation for the synchronization. If not
      provided, a default implementation is used.
    record_event_name: The name of the event to record the duration of the
      synchronization.
  """
  if should_skip_process_sync(processes):
    logging.vlog(
        1,
        '[process=%s][thread=%s] Skipping global process sync, barrier'
        ' name: %s',
        process_index(),
        threading.current_thread().name,
        name,
    )
    return
  timeout = timeout or _DEFAULT_BARRIER_TIMEOUT
  sync_start_time = time.time()
  # Temporarily default to existing behavior to minimize risk of breakage.
  if processes is None:
    key = _unique_barrier_key(name)
    logging.vlog(
        1,
        '[process=%s][thread=%s] Begin jax/sync_global_devices("%s")',
        process_index(),
        threading.current_thread().name,
        key,
    )
    multihost_utils.sync_global_devices(key)
    logging.vlog(
        1,
        '[process=%s][thread=%s] Done jax/sync_global_devices("%s"): %s secs',
        process_index(),
        threading.current_thread().name,
        key,
        (time.time() - sync_start_time),
    )
  else:
    barrier_sync_fn = barrier_sync_fn or get_barrier_sync_fn(
        processes=processes
    )
    barrier_sync_fn(key=name, timeout_ms=timeout * 1000)
  # This may end up just being too noisy given how many barriers there are, but
  # it does represent how long different processes waited around waiting for
  # other processes to reach a barrier.
  jax.monitoring.record_event_duration_secs(
      record_event_name,
      time.time() - sync_start_time,
  )


def _maybe_log_reached_preemption(
    step: int, preemption_sync_point_reached: bool
):
  if not preemption_sync_point_reached:
    return
  jax.monitoring.record_event('/jax/orbax/write/preemption')
  logging.warning(
      '[process=%s][thread=%s] Reached preemption sync point, step=%s',
      process_index(),
      threading.current_thread().name,
      step,
  )


def reached_preemption(step: int) -> bool:
  """Returns True if a preemption sync point has been reached."""

  preemption_sync_point_reached = multihost_utils.reached_preemption_sync_point(
      step
  )
  _maybe_log_reached_preemption(step, preemption_sync_point_reached)
  return preemption_sync_point_reached


def is_primary_host(primary_host: Optional[int]):
  if primary_host is None or primary_host == process_index():
    return True
  return False


def process_count() -> int:
  return jax.process_count()


def process_index() -> int:
  """Customized logic for obtaining JAX process index."""
  try:
    experimental_orbax_use_distributed_process_id = (
        EXPERIMENTAL_ORBAX_USE_DISTRIBUTED_PROCESS_ID.value
    )
  except Exception:  # pylint: disable=broad-exception-caught
    logging.log_first_n(
        logging.INFO,
        '[thread=%s] Failed to get flag value for'
        ' EXPERIMENTAL_ORBAX_USE_DISTRIBUTED_PROCESS_ID.',
        1,
        threading.current_thread().name,
    )
    experimental_orbax_use_distributed_process_id = False
  if experimental_orbax_use_distributed_process_id:
    logging.log_first_n(
        logging.INFO,
        '[thread=%s] Using distributed process id.',
        1,
        threading.current_thread().name,
    )
    return jax._src.distributed.global_state.process_id  # pylint: disable=protected-access
  else:
    return jax.process_index()


def unique_processes_from_devices(device_array: np.ndarray) -> Set[int]:
  get_pids_from_devices = np.vectorize(
      lambda d: runtime_to_distributed_process_id(d.process_index)
  )
  return set(get_pids_from_devices(device_array).flat)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines `SafetensorsLayout`, a class to handle Safetensors checkpoint formats."""

import json
import os
from typing import Any, Awaitable

import aiofiles
import jax
import numpy as np
from orbax.checkpoint._src.arrays import numpy_utils
from orbax.checkpoint.experimental.v1._src.layout import checkpoint_layout
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types

CheckpointLayout = checkpoint_layout.CheckpointLayout
InvalidLayoutError = checkpoint_layout.InvalidLayoutError
Path = types.Path

HEADER_NUM_BYTES = 8


def _get_dtypes() -> dict[str, Any]:
  """Returns the mapping from safetensor dtype strings to numpy dtypes."""
  return {
      "BOOL": np.bool_,
      "I8": np.int8,
      "U8": np.uint8,
      "I16": np.int16,
      "U16": np.uint16,
      "I32": np.int32,
      "U32": np.uint32,
      "I64": np.int64,
      "U64": np.uint64,
      "F16": np.float16,
      "F32": np.float32,
      "F64": np.float64,
      "F8_E8M0": "float8_e8m0fnu (specialized ML dtype)",
      "F4": "float4_e2m1fn_x2 (specialized ML dtype)",
  }


async def _read_safetensors_header(path: Path) -> tuple[dict[str, Any], int]:
  """Reads a safetensors file header, returning the header and data start offset."""
  async with aiofiles.open(path, mode="rb") as f:
    header_size_bytes = await f.read(HEADER_NUM_BYTES)
    if not header_size_bytes:
      raise ValueError("Could not read header size from safetensors file.")

    header_size = int.from_bytes(header_size_bytes, byteorder="little")
    header_bytes = await f.read(header_size)
    if len(header_bytes) != header_size:
      raise ValueError("Could not read header content from safetensors file.")

    header = json.loads(header_bytes)
    data_start_offset = HEADER_NUM_BYTES + header_size
    return header, data_start_offset


def _get_array_properties(info: dict[str, Any]) -> tuple[tuple[int, ...], Any]:
  """Parses shape and dtype from a safetensors tensor header."""
  try:
    dtype_str = info["dtype"]
    dtype = _get_dtypes()[dtype_str]
  except KeyError as e:
    raise ValueError(f"Unsupported dtype in SafeTensors header: {e}") from e
  shape = tuple(info["shape"])
  return shape, dtype


async def _read_non_contiguous_slice(
    f: aiofiles.threadpool.binary.AsyncBufferedIOBase,
    idx: tuple[slice, ...],
    stored_shape: tuple[int, ...],
    stored_dtype: np.dtype,
    tensor_file_offset: int,
) -> np.ndarray:
  """Reads a slice of a tensor from a file.

  This function solves the problem of reading a multi-dimensional slice from an
  array, where the slice's data is not stored as a single, contiguous block in 
  the file. It does so by recursively "walking" the dimensions of the slice.

  Args:
      f: The asynchronous file object (binary read mode)
      idx: A tuple of slice objects representing the n-dimensional slice to
        read.
      stored_shape: The shape of the tensor.
      stored_dtype: The dtype of the tensor.
      tensor_file_offset: The starting byte offset of the tensor's data within
        the file.

  Returns:
      The specific tensor slice.
  """
  # Handle 0-d scalar case
  if not idx:
    await f.seek(tensor_file_offset)
    num_bytes = np.dtype(stored_dtype).itemsize
    scalar_bytes = await f.read(num_bytes)
    # Reshape to () to create a 0-D NumPy array.
    return np.frombuffer(scalar_bytes, dtype=stored_dtype).reshape(())

  itemsize = np.dtype(stored_dtype).itemsize

  # Calculate the byte strides for the full tensor. The stride for a
  # dimension is the number of bytes to "jump" to get to the next element
  # in that dimension while keeping all other indices the same.
  global_strides = [itemsize] * len(stored_shape)
  for i in range(len(stored_shape) - 2, -1, -1):
    global_strides[i] = global_strides[i + 1] * stored_shape[i + 1]

  async def _read_slice_recursively(dim: int, base_offset: int) -> bytes:
    # TODO(b/438763866) - @zachmeyers to consider alternative methods.
    s = idx[dim]  # The slice for the current dimension.

    # If we are at the last dimension, the data is contiguous.
    if dim == len(stored_shape) - 1:
      start = base_offset + s.start * global_strides[dim]
      num_bytes = (s.stop - s.start) * itemsize
      await f.seek(tensor_file_offset + start)
      return await f.read(num_bytes)

    # For all other dimensions, iterate through the indices
    # of the slice and make a recursive call for the next dimension.
    chunks = []
    for i in range(s.start, s.stop):
      offset = base_offset + i * global_strides[dim]
      chunk = await _read_slice_recursively(dim + 1, offset)
      chunks.append(chunk)

    return b"".join(chunks)

  # Start the recursive reading process from the first dimension.
  slice_bytes = await _read_slice_recursively(dim=0, base_offset=0)
  shard_shape = numpy_utils.slice_shape(idx)
  return np.frombuffer(slice_bytes, dtype=stored_dtype).reshape(shard_shape)


async def _load_safetensors_fully_replicated(path: Path) -> dict[str, Any]:
  # TODO(b/438033067)
  """Reads a safetensors file and constructs the tensor dictionary."""
  header, data_start_offset = await _read_safetensors_header(path)
  tensors = {}
  async with aiofiles.open(path, mode="rb") as f:
    await f.seek(data_start_offset)
    data_bytes = await f.read()

  for name, info in header.items():
    if name == "__metadata__":
      continue
    shape, dtype = _get_array_properties(info)
    start_offset, end_offset = info["data_offsets"]
    tensor_bytes = data_bytes[start_offset:end_offset]
    np_array = np.frombuffer(tensor_bytes, dtype=dtype).reshape(shape)
    tensors[name] = np_array
  return {format_utils.PYTREE_CHECKPOINTABLE_KEY: tensors}


async def _load_safetensors_sharded(
    path: Path, abstract_pytree: dict[str, Any]
) -> dict[str, Any]:
  """Loads a safetensors file with sharding."""
  header, data_start_offset = await _read_safetensors_header(path)
  flat_abstract, _ = jax.tree.flatten_with_path(abstract_pytree)
  restored_pytree = {}

  async with aiofiles.open(path, mode="rb") as f:
    for key_path, abstract_leaf in flat_abstract:
      if len(key_path) != 1 or not isinstance(
          key_path[0], jax.tree_util.DictKey
      ):
        raise ValueError(
            f"The PyTree is not a flat dictionary. Key path: {key_path}"
        )
      tensor_name = str(key_path[0].key)
      if tensor_name not in header:
        raise KeyError(
            f"Tensor '{tensor_name}' not found in safetensors header."
        )

      stored_shape, stored_dtype = _get_array_properties(header[tensor_name])
      st_data_offsets = header[tensor_name]["data_offsets"]
      sharding = abstract_leaf.sharding
      target_shape = abstract_leaf.shape
      target_dtype = abstract_leaf.dtype

      device_indices_map = sharding.addressable_devices_indices_map(
          target_shape
      )

      device_map = []
      for device in device_indices_map:
        idx = device_indices_map[device]
        # Determine the slice corresponding to the device that the current
        # process is responsible for.
        resolved_idx = numpy_utils.resolve_slice(idx, stored_shape)
        shard_shape = numpy_utils.slice_shape(resolved_idx)

        shard_np = await _read_non_contiguous_slice(
            f,
            resolved_idx,
            stored_shape,
            stored_dtype,
            st_data_offsets[0] + data_start_offset,
        )

        shard_np = shard_np.reshape(shard_shape)

        if shard_np.dtype != target_dtype:
          shard_np = shard_np.astype(target_dtype)

        device_map.append(jax.device_put(shard_np, device))

      restored_pytree[tensor_name] = jax.make_array_from_single_device_arrays(
          target_shape, sharding, device_map
      )

  return {format_utils.PYTREE_CHECKPOINTABLE_KEY: restored_pytree}


class SafetensorsLayout(CheckpointLayout):
  """SafetensorsLayout.

  This class defines a class to handle Safetensors checkpoint formats. It
  inherits
  abstract methods from CheckpointLayout. It performs a few core functions:
    - Resolves handlers for saving and loading.
    - Saves and loads checkpointables to/from individual subdirectories by
    delegating to the resolved handlers.
  """

  def __init__(self, path: Path):
    self._path = path

  @property
  def path(self) -> Path:
    """Returns the path of the SafeTensors checkpoint."""
    return self._path

  async def metadata(self) -> metadata_types.CheckpointMetadata[dict[str, Any]]:
    """Returns the metadata of the SafeTensors checkpoint."""
    header, _ = await _read_safetensors_header(self._path)

    metadata = {}
    for name, info in header.items():
      if name == "__metadata__":
        continue
      shape, dtype = _get_array_properties(info)
      metadata[name] = jax.ShapeDtypeStruct(shape=shape, dtype=dtype)

    custom_metadata = header.get("__metadata__")
    commit_timestamp_nsecs = int(os.stat(self._path).st_mtime)

    return metadata_types.CheckpointMetadata[dict[str, Any]](
        metadata={format_utils.PYTREE_CHECKPOINTABLE_KEY: metadata},
        commit_timestamp_nsecs=commit_timestamp_nsecs,
        custom_metadata=custom_metadata,
    )

  def validate(self):
    if self._path.is_file() and self._path.suffix == ".safetensors":
      return
    else:
      raise InvalidLayoutError(
          f"Failed to interpret path {self._path} as a SafeTensors checkpoint."
          " A SafeTensors checkpoint must be a file with the '.safetensors'"
          " suffix."
      )

  def validate_pytree(self, checkpointable_name: str | None) -> None:
    return

  async def load(
      self,
      abstract_checkpointables: dict[str, Any] | None = None,
  ) -> Awaitable[dict[str, Any]]:
    abstract_pytree = None
    if abstract_checkpointables:
      abstract_pytree = abstract_checkpointables.get(
          format_utils.PYTREE_CHECKPOINTABLE_KEY
      )
    if abstract_pytree is not None:
      return _load_safetensors_sharded(self._path, abstract_pytree)
    else:
      return _load_safetensors_fully_replicated(self._path)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines free-function interface for partial saving and finalizing."""

import asyncio

from etils import epath
from orbax.checkpoint._src.path import async_path
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
import orbax.checkpoint.experimental.v1._src.handlers.global_registration  # pylint: disable=unused-import
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.saving import saving as saving_lib
from orbax.checkpoint.experimental.v1._src.synchronization import multihost
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types


PYTREE_CHECKPOINTABLE_KEY = saving_lib.PYTREE_CHECKPOINTABLE_KEY
PARTIAL_SAVE_SUFFIX = '.partial_save'


def _is_partial_save_path(path: path_types.PathLike) -> bool:
  path = epath.Path(path)
  return path.name.endswith(PARTIAL_SAVE_SUFFIX)


def _add_partial_save_suffix(path: path_types.PathLike) -> path_types.Path:
  path = epath.Path(path)
  return path.parent / (path.name + PARTIAL_SAVE_SUFFIX)


def _remove_partial_save_suffix(path: path_types.PathLike) -> path_types.Path:
  path = epath.Path(path)
  return path.parent / path.name.removesuffix(PARTIAL_SAVE_SUFFIX)


def save_pytree(
    path: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    custom_metadata: tree_types.JsonType | None = None,
):
  """Partially saves a PyTree.

  This function allows for incrementally updating a checkpoint. It is designed
  to be called multiple times. The first call initiates a new partial save
  "session" in a temporary location. Subsequent calls will update this session
  by modifying the checkpoint in place.

  The operation is atomic; if it is interrupted, the previous version of the
  partial save will be preserved.

  IMPORTANT: The checkpoint is not finalized at the target `path` until
  `ocp.partial.finalize(path)` is called. The intermediate checkpoints are
  temporary and should not be used directly.

  ### Workflow

  A typical partial save workflow involves one or more calls to
  `partial.save_pytree` followed by a single call to `partial.finalize`.

  ```
  path = '/path/to/my/checkpoint'

  # The first call creates a temporary directory:
  # '/path/to/my/checkpoint.partial_save'
  # Note: the exact temporary directory name is an implementation detail that
  # depends on the file system and should not be relied on.
  ocp.partial.save_pytree(path, {'layer1': ..., 'step': 1})

  # A subsequent call reads the previous version and applies new updates
  # to the temporary directory:
  # '/path/to/my/checkpoint.partial_save'
  ocp.partial.save_pytree(path, {'layer2': ..., 'metrics': ...})

  # This call commits the latest version to the final destination at
  # '/path/to/my/checkpoint'.
  ocp.partial.finalize(path)
  ```

  ### Additions vs. Replacements

  The provided `pytree` represents a set of updates.
  - If a key in `pytree` (e.g., 'metrics') does not exist in the on-disk
    checkpoint, it is treated as an **addition**. In other words, the sets of
    keys of the on-disk PyTree and the provided `pytree` are disjoint.
  - If a key (e.g., 'step') already exists, its value is **replaced**. In other
    words, the sets of keys of the on-disk PyTree and the provided `pytree`
    overlap. Replacements are currently NOT supported. Please reach out to the
    Orbax team if you need this functionality.

  See `ocp.save_pytree` for general PyTree saving documentation.

  Args:
    path: The path to save the checkpoint to.
    pytree: A PyTree representing the additions to be applied to the on-disk
      checkpoint.
    custom_metadata: User-provided custom metadata. This will be merged with any
      existing custom metadata. Values from this dictionary will overwrite
      existing values if keys conflict.
  """
  save_pytree_async(
      path,
      pytree,
      custom_metadata=custom_metadata,
  ).result()


def save_pytree_async(
    path: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    custom_metadata: tree_types.JsonType | None = None,
) -> async_types.AsyncResponse[None]:
  """Partially saves a PyTree asynchronously.

  Unlike `partial.save_pytree`, this function returns an `AsyncResponse`
  immediately after scheduling the save operation. The actual writing to disk
  happens in a background thread. You can use `response.result()` to block
  until the operation is complete.

  This function allows for incrementally updating a checkpoint. It is designed
  to be called multiple times. The first call initiates a new partial save
  "session" in a temporary location. Subsequent calls will update this session
  by creating a new version that includes all previous changes plus the new
  ones.

  The operation is atomic; if it is interrupted, the previous version of the
  partial save will be preserved.

  IMPORTANT: The checkpoint is not finalized at the target `path` until
  `ocp.partial.finalize(path)` is called. The intermediate checkpoints are
  temporary and may be garbage collected in certain environments.

  ### Workflow

  A typical partial save workflow involves one or more calls to
  `partial.save_pytree_async` followed by a single call to `partial.finalize`.

  ```
  path = '/path/to/my/checkpoint'

  # The first call creates a temporary directory and returns immediately.
  response1 = ocp.partial.save_pytree_async(path, {'layer1': ..., 'step': 1})

  # A subsequent call also returns immediately. Orbax ensures that this
  # operation waits for the first one to complete before starting.
  response2 = ocp.partial.save_pytree_async(
      path, {'layer2': ..., 'metrics': ...}
  )

  # Wait for all async partial saves to complete before finalizing.
  response1.result()
  response2.result()

  # This call commits the latest version to the final destination at
  # '/path/to/my/checkpoint'.
  ocp.partial.finalize(path)
  ```

  ### Additions vs. Replacements

  The provided `pytree` represents a set of updates.
  - If a key in `pytree` (e.g., 'metrics') does not exist in the on-disk
    checkpoint, it is treated as an **addition**.
  - If a key (e.g., 'step') already exists, its value is **replaced**.
    Replacements are currently NOT supported. Please reach out to the Orbax team
    if you need this functionality.

  See `ocp.save_pytree_async` for general PyTree saving documentation.

  Args:
    path: The path to save the checkpoint to.
    pytree: The PyTree to save. This may be any JAX PyTree (including custom
      objects registered as PyTrees) consisting of supported leaf types. Default
      supported leaf types include `jax.Array`, `np.ndarray`, simple types like
      `int`, `float`, `str`, and empty nodes. Support for custom leaves is also
      possible by implementing a `LeafTypeHandler`.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.

  Returns:
    An `AsyncResponse` that can be used to block until the save is complete.
    Blocking can be done using `response.result()`, which returns `None`.
  """
  return saving_lib._save_checkpointables_impl(  # pylint: disable=protected-access
      path,
      {saving_lib.PYTREE_CHECKPOINTABLE_KEY: pytree},
      overwrite=False,
      custom_metadata=custom_metadata,
      async_origin=True,
      # partial_save=True,
  )


def finalize(path: path_types.PathLike) -> None:
  """Finalizes a partially-saved checkpoint, making it permanent and readable.

  This function commits all changes made during a partial save session,
  concluding the transaction. It should be called once after all desired
  `ocp.partial.save_*` operations are complete.

  The finalization process is atomic. It renames the temporary, versioned
  partial save directory to the final target `path`, making the updated
  checkpoint "live".

  IMPORTANT: Until `finalize` is called, the checkpoint at the target `path`
  is not created or modified. All changes are buffered in a temporary location.
  This function is what makes those changes permanent.


  ### Example
  ```
  path = '/path/to/my/checkpoint'

  # These calls write to a temporary, versioned directory, not the final path.
  ocp.partial.save_pytree(path, {'step': 1})
  ocp.partial.save_checkpointables(path, {'metrics': ...})

  # This call performs the atomic rename, making the checkpoint available at
  # '/path/to/my/checkpoint'.
  ocp.partial.finalize(path)
  ```

  Args:
    path: The final, target path of the checkpoint to be finalized. This should
      be the same path that was passed to `ocp.partial.save_*` calls.

  Raises:
    FileExistsError: If a finalized checkpoint already exists at `path`. To
      overwrite, it must be deleted first.
    FileNotFoundError: If no partial save session is found for the given `path`.
      This can happen if `ocp.partial.save_*` was not called first.
  """
  context = context_lib.get_context()

  path = epath.Path(path)
  if _is_partial_save_path(path):
    final_path = _remove_partial_save_suffix(path)
    partial_path = path
  else:
    final_path = path
    partial_path = _add_partial_save_suffix(path)

  async def _finalize_impl():
    if await async_path.exists(final_path):
      raise FileExistsError(
          f'Finalized checkpoint already exists at {final_path}.'
      )
    elif not await async_path.exists(partial_path):
      raise FileNotFoundError(
          f'Partial save path {partial_path} does not exist.'
      )

    await multihost.sync_global_processes(
        multihost.unique_barrier_key(
            'OcpPartialSaving:finalize_error',
            prefix=context.multiprocessing_options.barrier_sync_key_prefix,
        ),
        processes=context.multiprocessing_options.active_processes,
    )

    rename_failed = False
    rename_error = None
    if multihost.is_primary_host(context.multiprocessing_options.primary_host):
      try:
        await async_path.rename(partial_path, final_path)
      except OSError as e:
        rename_failed = True
        rename_error = e

    rename_failed = multihost.broadcast_one_to_all(
        rename_failed,
        is_source=multihost.is_primary_host(
            context.multiprocessing_options.primary_host
        ),
    )

    if rename_failed:
      if rename_error is not None:
        raise rename_error
      raise OSError('Partial checkpoint finalization failed during rename.')

  asyncio.run(_finalize_impl())



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Registry for checkpoint layouts."""

from etils import epath
from orbax.checkpoint.experimental.v1._src.context import options as options_lib
from orbax.checkpoint.experimental.v1._src.layout import checkpoint_layout
from orbax.checkpoint.experimental.v1._src.layout import orbax_layout
from orbax.checkpoint.experimental.v1._src.layout import safetensors_layout
from orbax.checkpoint.experimental.v1._src.path import types as path_types

InvalidLayoutError = checkpoint_layout.InvalidLayoutError


def get_checkpoint_layout(
    path: path_types.PathLike, layout_enum: options_lib.CheckpointLayout
) -> checkpoint_layout.CheckpointLayout:
  """Returns the checkpoint layout class for the given path.

  Args:
    path: The path to the checkpoint.
    layout_enum: The checkpoint layout to use.

  Returns:
    The class of the matching CheckpointLayout.

  Raises:
    InvalidLayoutError: If the path is not a valid checkpoint for any registered
    layout, with details from each layout's validation attempt.
  """
  path = epath.Path(path)

  match layout_enum:
    case options_lib.CheckpointLayout.ORBAX:
      layout_class = orbax_layout.OrbaxLayout
    case options_lib.CheckpointLayout.SAFETENSORS:
      layout_class = safetensors_layout.SafetensorsLayout
    case _:
      raise ValueError(f"Unsupported checkpoint layout: {layout_enum}")

  try:
    layout = layout_class(path)
    layout.validate()
    return layout
  except InvalidLayoutError as e:
    raise InvalidLayoutError(
        f"Could not recognize the checkpoint at {path} as a valid"
        f" {layout_enum.value} checkpoint. If you are trying to load a"
        " checkpoint that does not conform to the stadnard Orbax format, use"
        " `ocp.Context(layout=...)` to specify the expected checkpoint layout."
    ) from e



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines free-function interface for saving."""

import asyncio
import time
from typing import Any

from etils import epath
import nest_asyncio
from orbax.checkpoint._src.checkpointers import async_checkpointer
from orbax.checkpoint._src.handlers import composite_checkpoint_handler
from orbax.checkpoint._src.handlers import handler_registration as legacy_handler_registration
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import compatibility as handler_compatibility
from orbax.checkpoint.experimental.v1._src.handlers import registration as handler_registration
import orbax.checkpoint.experimental.v1._src.handlers.global_registration  # pylint: disable=unused-import
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types as path_types
from orbax.checkpoint.experimental.v1._src.saving import saving_utils
from orbax.checkpoint.experimental.v1._src.synchronization import types as async_types
from orbax.checkpoint.experimental.v1._src.tree import types as tree_types

PYTREE_CHECKPOINTABLE_KEY = format_utils.PYTREE_CHECKPOINTABLE_KEY


def save_pytree(
    path: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
):
  """Saves a PyTree.

  The operation blocks until complete. For improved performance, consider using
  `save_async` instead.

  Args:
    path: The path to save the checkpoint to.
    pytree: The PyTree to save. This may be any JAX PyTree (including custom
      objects registered as PyTrees) consisting of supported leaf types. Default
      supported leaf types include `jax.Array`, `np.ndarray`, simple types like
      `int`, `float`, `str`, and empty nodes. Support for custom leaves is also
      possible by implementing a `LeafTypeHandler`.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """
  save_checkpointables(
      path,
      {PYTREE_CHECKPOINTABLE_KEY: pytree},
      overwrite=overwrite,
      custom_metadata=custom_metadata,
  )


def save_checkpointables(
    path: path_types.PathLike,
    checkpointables: dict[str, Any],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
) -> None:
  """Saves a dictionary of checkpointables.

  A “checkpointable” refers to a logical piece of the checkpoint that is
  distinct in some way from other pieces. Checkpointables are separable;
  they may or may not be loaded concurrently and some may be omitted from the
  checkpoint entirely. Checkpointables are often represented by different types,
  and have different representations on disk. The quintessential example is
  model params vs. dataset.

  For example, one might do::

    ocp.save_checkpointables(
        path,
        {
            'params': pytree_of_arrays,
            'dataset': pygrain.DatasetIterator(...),
        }
    )

  It is also possible to do::

    train_state = {
        'params': params_pytree_of_arrays,
        'opt_state': opt_state_pytree_of_arrays,
        'step': step,
        ...
    }
    ocp.save_checkpointables(path, train_state)

  This is not the ideal way of doing things because it is then difficult to run
  transformations that involve the entire train state (see the
  `load_and_transform` API).

  Args:
    path: The path to save the checkpoint to.
    checkpointables: A dictionary of checkpointables. Dictionary keys represent
      the names of the checkpointables, while the values are the checkpointable
      objects themselves.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.
  """
  _save_checkpointables_impl(
      path,
      checkpointables,
      overwrite=overwrite,
      custom_metadata=custom_metadata,
      async_origin=False,
  ).result()


# TODO(b/396190818): Test modification of the context by the user after the
# save operation is scheduled.
def save_pytree_async(
    path: path_types.PathLike,
    pytree: tree_types.PyTreeOf[tree_types.LeafType],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
) -> async_types.AsyncResponse[None]:
  """Saves a PyTree asynchronously.

  Unlike `save_pytree`, this function returns immediately after the save
  operation is scheduled
  (except for certain operations, like device-to-host copying of
  on-device arrays, which must happen on the main thread). Further writing
  operations continue in a background thread. An `AsyncResponse` is returned
  that can be used to block until the save is complete (using
  `response.result()`). Make sure to wait for completion before attempting to
  load the checkpoint or exiting the program.

  Args:
    path: The path to save the checkpoint to.
    pytree: The PyTree to save. This may be any JAX PyTree (including custom
      objects registered as PyTrees) consisting of supported leaf types. Default
      supported leaf types include `jax.Array`, `np.ndarray`, simple types like
      `int`, `float`, `str`, and empty nodes. Support for custom leaves is also
      possible by implementing a `LeafTypeHandler`.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.

  Returns:
    An `AsyncResponse` that can be used to block until the save is complete.
    Blocking can be done using `response.result()`, which returns `None`.
  """
  return save_checkpointables_async(
      path,
      {PYTREE_CHECKPOINTABLE_KEY: pytree},
      overwrite=overwrite,
      custom_metadata=custom_metadata,
  )


def save_checkpointables_async(
    path: path_types.PathLike,
    checkpointables: dict[str, Any],
    *,
    overwrite: bool = False,
    custom_metadata: tree_types.JsonType | None = None,
) -> async_types.AsyncResponse[None]:
  """Saves a dictionary of checkpointables asynchronously.

  See `save_checkpointables` documentation.

  Unlike `save_checkpointables`, this function returns immediately after the
  save operation is scheduled
  (except for certain operations, like device-to-host copying of
  on-device arrays, which must happen on the main thread). Further writing
  operations continue in a background thread. An `AsyncResponse` is returned
  that can be used to block until the save is complete (using
  `response.result()`). Make sure to wait for completion before attempting to
  load the checkpoint or exiting the program.

  Args:
    path: The path to save the checkpoint to.
    checkpointables: A dictionary of checkpointables. Dictionary keys represent
      the names of the checkpointables, while the values are the checkpointable
      objects themselves.
    overwrite: If True, fully overwrites an existing checkpoint in `path`.
      Otherwise, raises an error if the checkpoint already exists.
    custom_metadata: User-provided custom metadata. An arbitrary
      JSON-serializable dictionary the user can use to store additional
      information. The field is treated as opaque by Orbax.

  Returns:
    An `AsyncResponse` that can be used to block until the save is complete.
    Blocking can be done using `response.result()`, which returns `None`.
  """
  return _save_checkpointables_impl(
      path,
      checkpointables,
      overwrite=overwrite,
      custom_metadata=custom_metadata,
      async_origin=True,
  )


def _save_checkpointables_impl(
    path: path_types.PathLike,
    checkpointables: dict[str, Any],
    *,
    async_origin: bool,
    overwrite: bool,
    custom_metadata: tree_types.JsonType | None,
) -> async_types.AsyncResponse[None]:
  """See caller docstrings."""
  nest_asyncio.apply()
  context = context_lib.get_context()
  path = epath.Path(path)
  # Prevent internal mutation from affecting the caller.
  checkpointables = dict(checkpointables)

  start_time = time.time()
  saving_utils.record_save_start(path, async_origin=async_origin)

  tmp_path = saving_utils.get_temporary_path(path, context=context)

  checkpointables = saving_utils.add_internal_checkpointables(
      checkpointables, context=context
  )

  background_awaitable = asyncio.run(
      saving_utils.run_blocking_save(
          tmp_path,
          checkpointables,
          overwrite=overwrite,
          context=context,
      )
  )

  return saving_utils.create_save_response(
      background_awaitable,
      checkpointables,
      tmp_path,
      start_time,
      context=context,
      custom_metadata=custom_metadata,
      async_origin=async_origin,
  )


def get_v0_checkpointer_and_args(
    checkpointables: dict[str, Any],
    *,
    metrics: tree_types.JsonType | None = None,
    context: context_lib.Context,
) -> tuple[
    async_checkpointer.AsyncCheckpointer,
    composite_checkpoint_handler.CompositeArgs,
]:
  """Construct V0 Checkpointer and Args for saving."""
  if (
      provided_reserved_keys := checkpointables.keys()
      & format_utils.RESERVED_CHECKPOINTABLE_KEYS
  ):
    raise ValueError(
        f'Provided reserved checkpointable keys: {provided_reserved_keys}.'
    )
  checkpointables = saving_utils.add_internal_checkpointables(
      checkpointables, context=context, metrics=metrics
  )

  handlers = {
      name: handler_registration.resolve_handler_for_save(
          context.checkpointables_options.registry, checkpointable, name=name
      )
      for name, checkpointable in checkpointables.items()
  }
  compatibility_handlers = {
      name: handler_compatibility.get_compatibility_handler(handler)
      for name, handler in handlers.items()
  }
  handler_registry = (
      legacy_handler_registration.DefaultCheckpointHandlerRegistry()
  )
  for name, handler in compatibility_handlers.items():
    handler_registry.add(name, handler_compatibility.Args, handler)
  composite_options = composite_checkpoint_handler.CompositeOptions(
      async_options=context.async_options.v0(),
      file_options=context.file_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  ckptr = async_checkpointer.AsyncCheckpointer(
      composite_checkpoint_handler.CompositeCheckpointHandler(
          handler_registry=handler_registry,
          composite_options=composite_options,
      ),
      async_options=context.async_options.v0(),
      multiprocessing_options=context.multiprocessing_options.v0(),
      file_options=context.file_options.v0(),
      temporary_path_class=context.file_options.temporary_path_class,
  )
  args = composite_checkpoint_handler.CompositeArgs(**{
      name: handler_compatibility.Args(checkpointable)
      for name, checkpointable in checkpointables.items()
  })
  return ckptr, args



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Defines `OrbaxLayout`, a class to handle Orbax checkpoint formats."""
from typing import Any, Awaitable
from orbax.checkpoint.experimental.v1._src.context import context as context_lib
from orbax.checkpoint.experimental.v1._src.handlers import composite_handler
from orbax.checkpoint.experimental.v1._src.handlers import registration
from orbax.checkpoint.experimental.v1._src.layout import checkpoint_layout
from orbax.checkpoint.experimental.v1._src.loading import v0_compatibility
from orbax.checkpoint.experimental.v1._src.metadata import types as metadata_types
from orbax.checkpoint.experimental.v1._src.path import format_utils
from orbax.checkpoint.experimental.v1._src.path import types

InvalidLayoutError = checkpoint_layout.InvalidLayoutError
CompositeHandler = composite_handler.CompositeHandler
Path = types.Path
CheckpointLayout = checkpoint_layout.CheckpointLayout

ORBAX_CHECKPOINT_INDICATOR_FILE = (
    composite_handler.ORBAX_CHECKPOINT_INDICATOR_FILE
)


_V0_ERROR_MESSAGE = (
    "If your checkpoint was saved with the Orbax V0 API, please follow the"
    " instructions at"
    " https://orbax.readthedocs.io/en/latest/guides/checkpoint/v1/orbax_v0_to_v1_migration.html"
    " to load it with the Orbax V1 API."
)
_GENERAL_ERROR_MESSAGE = (
    " Note that a valid checkpoint path should always contain a file named"
    f" '{ORBAX_CHECKPOINT_INDICATOR_FILE}' (unless it was saved with the V0"
    f" API). {_V0_ERROR_MESSAGE}"
)


class OrbaxLayout(CheckpointLayout):
  """OrbaxLayout.

  This class defines a class to handle Orbax checkpoint formats. It inherits
  abstract methods from CheckpointLayout. It performs a few core functions:
    - Resolves handlers for saving and loading.
    - Saves and loads checkpointables to/from individual subdirectories by
    delegating to the resolved handlers.
  """

  def __init__(self, path: Path):
    self._context = context_lib.get_context()
    self._handler_registry = registration.local_registry(
        self._context.checkpointables_options.registry,
        include_global_registry=False,
    )
    self._composite_handler = CompositeHandler(self._handler_registry)
    self._path = path

  @property
  def path(self) -> Path:
    """Returns the path of the Orbax checkpoint."""
    return self._path

  @property
  def has_indicator_file(self) -> bool:
    return (self._path / ORBAX_CHECKPOINT_INDICATOR_FILE).exists()

  async def metadata(self) -> metadata_types.CheckpointMetadata[dict[str, Any]]:
    # Uses the v0 checkpointer to get v0 StepMetadata
    checkpointer, _ = v0_compatibility.get_v0_checkpointer_and_args(
        self._path, None, context=context_lib.get_context()
    )
    step_metadata = checkpointer.metadata(self._path)

    item_metadata = {k: v for k, v in step_metadata.item_metadata.items()}
    # Exclude `metrics` if present. This is relevant only for
    # `training.Checkpointer`, and is separately added to the
    # `training.CheckpointMetadata` object.
    item_metadata.pop("metrics", None)

    return metadata_types.CheckpointMetadata[dict[str, Any]](
        metadata=item_metadata,
        init_timestamp_nsecs=step_metadata.init_timestamp_nsecs,
        commit_timestamp_nsecs=step_metadata.commit_timestamp_nsecs,
        custom_metadata=step_metadata.custom_metadata,
    )

  def validate(self):
    try:
      format_utils.validate_checkpoint_directory(self._path)
      if self.has_indicator_file:
        format_utils.validate_checkpoint_metadata(self._path)
    except BaseException as e:
      raise InvalidLayoutError(
          f"Failed to interpret path {self._path} as an Orbax checkpoint."
          f" {_GENERAL_ERROR_MESSAGE}"
      ) from e

  def validate_pytree(self, checkpointable_name: str | None) -> None:
    """Validates the given path as a PyTree checkpoint."""
    try:
      format_utils.validate_pytree_checkpoint(
          self._path, checkpointable_name=checkpointable_name
      )
    except BaseException as e:
      raise InvalidLayoutError(
          f"Failed to interpret path {self._path} as an Orbax PyTree"
          f" checkpoint. {_GENERAL_ERROR_MESSAGE}"
      ) from e

  async def load(
      self,
      abstract_checkpointables: dict[str, Any] | None = None,
  ) -> Awaitable[dict[str, Any]]:
    load_awaitable = await self._composite_handler.load(
        self._path, abstract_checkpointables
    )
    return load_awaitable



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Snapshot represents operations on how to create, delete snapshots of a checkpoint."""

import abc
import asyncio
import time

from absl import logging
from etils import epath
from orbax.checkpoint._src.path import utils as ocp_path_utils


SNAPSHOTTING_TIME = "snapshotting_time"


class Snapshot(abc.ABC):
  """Represents a snapshot of a checkpoint."""

  _source: epath.Path
  _snapshot: epath.Path

  @abc.abstractmethod
  async def create_snapshot(self) -> None:
    """Creates a snapshot of the checkpoint."""
    pass

  @abc.abstractmethod
  async def release_snapshot(self) -> None:
    """Deletes a snapshot of the checkpoint."""
    pass

  @abc.abstractmethod
  async def replace_source(self) -> None:
    """Replaces the source checkpoint with the snapshot."""
    pass


class _DefaultSnapshot(Snapshot):
  """Creates a copy of the checkpoint in the snapshot folder."""

  def __init__(
      self,
      src: epath.PathLike,
      dst: epath.PathLike,
  ):
    self._source = epath.Path(src)
    self._snapshot = epath.Path(dst)

  async def create_snapshot(self) -> None:
    """Creates a deep copy of the checkpoint."""
    if not await asyncio.to_thread(self._snapshot.is_absolute):
      raise ValueError(
          f"Snapshot destination must be absolute, but was '{self._snapshot}'."
      )
    if not await asyncio.to_thread(self._source.exists):
      raise ValueError(f"Snapshot source does not exist: {self._source}'.")

    t = ocp_path_utils.Timer()
    await asyncio.to_thread(
        ocp_path_utils.recursively_copy_files,
        self._source,
        self._snapshot,
    )
    logging.debug(
        "Snapshot copy: %fs",
        t.get_duration(),
    )

  async def release_snapshot(self) -> None:
    """Deletes a snapshot of the checkpoint."""
    if not await asyncio.to_thread(self._snapshot.exists):
      raise FileNotFoundError(f"Snapshot does not exist: {self._snapshot}")

    await asyncio.to_thread(self._snapshot.rmtree)

  # TODO(b/434025182): Handle recovery path upon restart.
  async def replace_source(self) -> None:
    """Replaces the source checkpoint with the snapshot."""
    if not self._snapshot.is_absolute():
      raise ValueError(
          f"Snapshot destination must be absolute, but was '{self._snapshot}'."
      )
    if not self._source.is_absolute():
      raise ValueError(
          f"Snapshot source must be absolute, but was '{self._source}'."
      )

    recovery_path = (
        self._source.parent / f"{self._source.name}._recovery_{time.time()}"
    )

    def _swap_source_and_snapshot():
      self._source.rename(recovery_path)
      self._snapshot.rename(self._source)
      recovery_path.rmtree()

    await asyncio.to_thread(_swap_source_and_snapshot)




def create_instance(
    source: epath.Path,
    snapshot: epath.Path,
    *,
    set_immutable: bool | None = None,
):
  return _DefaultSnapshot(source, snapshot)



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Validation functions involved in loading."""

from orbax.checkpoint.experimental.v1._src.path import format_utils


def validate_abstract_checkpointables(abstract_checkpointables):
  if abstract_checkpointables is None:
    return
  if (
      provided_reserved_keys := abstract_checkpointables.keys()
      & format_utils.RESERVED_CHECKPOINTABLE_KEYS
  ):
    raise ValueError(
        f'Provided reserved checkpointable keys: {provided_reserved_keys}.'
    )



# Copyright 2025 The Orbax Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Utils for interacting with GCS paths."""

import functools
from urllib import parse
from etils import epath


def parse_gcs_path(path: epath.PathLike) -> tuple[str, str]:
  parsed = parse.urlparse(str(path))
  assert parsed.scheme == 'gs', f'Unsupported scheme for GCS: {parsed.scheme}'
  # Strip the leading slash from the path.
  standardized_path = parsed.path
  if standardized_path.startswith('/'):
    standardized_path = standardized_path[1:]
  # Add a trailing slash if it's missing.
  if not standardized_path.endswith('/'):
    standardized_path = standardized_path + '/'
  return parsed.netloc, standardized_path


@functools.lru_cache(maxsize=32)
def get_bucket(bucket_name: str):
  # pylint: disable=g-import-not-at-top
  from google.cloud import storage  # pytype: disable=import-error

  client = storage.Client()
  return client.get_bucket(bucket_name)


def is_hierarchical_namespace_enabled(path: epath.PathLike) -> bool:
  """Return whether hierarchical namespace is enabled."""
  bucket_name, _ = parse_gcs_path(path)
  bucket = get_bucket(bucket_name)
  return bucket.hierarchical_namespace_enabled
